#!/usr/bin/env python3

"""
data-storage that provides both attribute- and dict-like access
"""

# sardinecake - run GitLab-CI jobs within libvirt VMs
#
# Copyright © 2024, IOhannes m zmölnig, forum::für::umläute
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


class Namespace:
    """
    A data storage that provides both attribute- and dict-like access.

    """

    def __init__(self, **kwargs):
        self._dict = kwargs
        self.items = self._dict.items
        self.keys = self._dict.keys
        self.values = self._dict.values
        self.get = self._dict.get

    def __repr__(self):
        typename = type(self).__name__
        args = ", ".join(f"{k}: {v!r}" for k, v in self._dict.items())
        return f"{typename}({args})"

    def __len__(self):
        return self._dict.__len__()

    def __iter__(self):
        return self._dict.__iter__()

    def __getitem__(self, x):
        return self.get(x)

    def __getattr__(self, x):
        return self.get(x)

    def __contains__(self, x):
        return self._dict.__contains__(x)

    def __hasattr__(self, x):
        return self._dict.__contains__(x)

    def setdefaults(self, defaults: dict):
        """Insert each key with a value of defaults, if the key is not in the Namespace.

        Return the (updated) Namespace object.
        """
        for k, v in defaults.items():
            self._dict.setdefault(k, v)
        return self


if __name__ == "__main__":
    ns = Namespace(foo=12, bar="pizza")
    print(ns)

    def check(key, value, attrvalue):
        """assert"""
        if ns[key] != value:
            raise AssertionError(f"ns[{key!r}]={ns[key]!r} != {value!r}")
        if value != attrvalue:
            raise AssertionError(f"ns.{key}={attrvalue} != {value!r}")

    check("foo", 12, ns.foo)
    check("bar", "pizza", ns.bar)
    check("nada", None, ns.nada)



#!/usr/bin/env python3

# sardinecake - manage VMs via OCI
#
# Copyright © 2024, IOhannes m zmölnig, forum::für::umläute
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


# export a libvirt VM into an OCI bundle

import datetime
import hashlib
import json
import logging
import os
import tempfile
import time


import libvirt

try:
    from .domain import Domain
    from . import gziparchive as gz
    from . import qcow
    from . import utils
except ImportError:
    from domain import Domain
    import gziparchive as gz
    import qcow
    import utils


log = logging.getLogger("sardinecake.exportvm")


def export2json(data, outdir):
    """encodes <data> to JSON and writes it to outdir, using the hash of the content as filename.
    returns a (hash, filesize) tuple"""
    jdata = json.dumps(data, indent=2)
    name = hashlib.sha256(jdata.encode()).hexdigest()
    with open(os.path.join(outdir, name), "w") as f:
        size = f.write(jdata)
    return (name, size)


def checkRelativeDiskimages(disks: list) -> bool:
    errors = 0
    for disk in disks:
        try:
            imagechain = list(qcow.QCOW(disk).getBackingChain(normalize=False))
        except:
            log.exception("problems with getting the backing chain")
            imagechain = [disk]
        baseimage = imagechain[0]
        basedir = os.path.dirname(baseimage)
        backingimages = imagechain[1:]
        for idx, backimage in enumerate(backingimages):
            backdir = os.path.dirname(backimage)
            if not backdir:
                continue
            errors += 1
            log.error(
                f"backing file #{idx} for {baseimage!r} in {basedir!r} in the wrong directory!"
            )
            if os.path.abspath(os.path.join(basedir, backdir)) == os.path.abspath(
                basedir
            ):
                # backing file lives in the same directory as the original image,
                # but nevertheless uses a path
                # this can be easily fixed with 'qemu-img'
                log.error(
                    f"""You can make the backing file relative with the following command:
```
qemu-img rebase -F qcow2 -u -b {os.path.basename(backimage)!r} {os.path.abspath(baseimage)!r}
```"""
                )
            else:
                log.error(
                    f"""Disk image and backing file live in different directories
Use something like the following to fix this:
```
mv {os.path.abspath(os.path.join(basedir, backdir))!r} {os.path.abspath(basedir)!r}
qemu-img rebase -F qcow2 -u -b {os.path.basename(backimage)!r} {os.path.abspath(baseimage)!r}
```"""
                )
    return errors == 0


def _exportVM(
    vmname: str,
    outdir=".",
    libvirtURI: str | None = None,
    chunksize: int | None = None,
    vmconfig: dict = {},
    libvirtconfigfile: str | None = None,
    quiet: bool = False,
    allow_absolutebaseimages: bool = False,
) -> tuple[str, str] | None:
    """export a libvirt VM into an OCI compatible bundle.
    returns the SHA256 hash of the manifest file and the config file.
    or None (if something went wrong)
    """

    def chunkzip(filename, outdir, chunksize, mimetype=None, name=None):
        if mimetype is None:
            mimetype = utils.mimetype(filename)
        if mimetype:
            mimetype = f"{mimetype}+gzip"
        chunks = gz.gzipchunked(filename, outdir, chunksize=chunksize, name=name)
        layers = []
        for c in chunks:
            chunk_stat = os.stat(os.path.join(outdir, c))
            layer = {
                "digest": f"sha256:{c}",
            }
            if mimetype:
                params = ""
                if layers:
                    params = f";chunk={len(layers)}"
                layer["mediaType"] = mimetype + params
            layer["size"] = chunk_stat.st_size

            layers.append(layer)
        return layers

    def timestamp2iso(timestamp):
        ts = datetime.datetime.fromtimestamp(
            timestamp, datetime.datetime.now().astimezone().tzinfo
        )
        return ts.isoformat()

    mtime = None
    newimages = {}
    newdisks = {}
    layers = []
    history = []
    diffids = []

    config = {}
    image_cfg = {}

    if libvirtconfigfile:
        domain = Domain.fromXMLFile(libvirtconfigfile)
    else:
        domain = Domain.fromLibVirt(libvirtURI, vmname, quiet=quiet)
    if not domain:
        log.fatal(f"VM {vmname} not found!")
        return

    image_cfg = {
        "architecture": domain.getCPU(),
        "os": domain.getOS(),
        "created": None,
    }

    disks = domain.getDiskFiles()
    if not disks:
        raise KeyError(f"VM {vmname!r} does not have any exportable disks")

    # check if disk images are good
    if not checkRelativeDiskimages(disks.values()):
        if allow_absolutebaseimages:
            log.warning("Proceeding despite possible problems with disk images")
        else:
            log.fatal("Refusing to export with problematic disk images")
            return

    os.makedirs(outdir, exist_ok=True)

    # layer #1: the libvirt VM definition (added at the end)

    # layer #2: vm sardine config (username, password,...)
    if vmconfig:
        vmcfg = export2json(vmconfig, outdir)
        layers += [
            {
                "digest": f"sha256:{vmcfg[0]}",
                "mediaType": "application/vnd.sardinecake.vm.config.v1+json",
                "size": vmcfg[1],
            }
        ]
        history += [
            {
                "created": timestamp2iso(time.time()),
            }
        ]
        diffids.append(f"sha256:{vmcfg[0]}")

    # layer #3..N: disks
    for dev, diskimage in disks.items():
        try:
            imagechain = reversed(qcow.QCOW(diskimage).getBackingChain())
        except:
            log.debug(f"couldn't get backing chain for {diskimage!r}", exc_info=True)
            imagechain = [diskimage]

        basename = None
        for d in imagechain:
            basename = os.path.basename(d)
            n, x = os.path.splitext(basename)
            infix = ""
            while f"{n}{infix}{x}" in newimages:
                if not infix:
                    infix = 0
                infix -= 1
                basename = f"{n}{infix}{x}"
                log.error(f"{d!r} already exists...adding to {basename}")

            if not d:
                log.error(f"incomplete backing chain for {diskimage!r}")
                continue
            # store the hash of the uncompressed data
            log.info(f"hashing {d!r}")
            with open(d, "rb") as f:
                digest = hashlib.file_digest(f, "sha256")
                diffids.append(f"sha256:{digest.hexdigest()}")
                history += [
                    {
                        "created": time.time(),  # os.stat(d).st_mtime,
                        "created_by": f"gzip -6 -c {os.path.basename(d)} | split -b {chunksize} {os.path.basename(d)}.gz # approximately",
                    }
                ]
            log.info(f"chunkzipping {d!r} as {basename}")
            layers += chunkzip(d, outdir, chunksize=chunksize, name=basename)
            newimages[basename] = True
        if basename:
            newdisks[dev] = basename

    domain.changeDiskFiles(newdisks)
    domain.anonymize()

    # layer #1: VMconfig
    vmconfig = domain.toXML()
    vmconfig_hash = hashlib.sha256(vmconfig.encode()).hexdigest()
    with open(os.path.join(outdir, vmconfig_hash), "w") as f:
        vmconfig_size = f.write(vmconfig)
        layers.insert(
            0,
            {
                "digest": f"sha256:{vmconfig_hash}",
                "mediaType": "application/vnd.libvirt.domain.config+xml",  # isn't there a real mimetype?
                "size": vmconfig_size,
            },
        )
        history.insert(
            0,
            {
                "created": timestamp2iso(time.time()),
                "created_by": f"virsh dumpxml {vmname!r}",
            },
        )
        diffids.insert(
            0,
            f"sha256:{vmconfig_hash}",
        )

    # mtime = max(_["mtime"] for _ in alldisks)
    if diffids:
        image_cfg["rootfs"] = {
            "type": "layers",  # not really true!
            "diff_ids": diffids,
        }

    if history:
        image_cfg["history"] = history

    if mtime:
        mtimestamp = datetime.datetime.fromtimestamp(
            mtime, datetime.datetime.now().astimezone().tzinfo
        )
        image_cfg["created"] = timestamp2iso(mtime)
    else:
        image_cfg["created"] = timestamp2iso(time.time())

    if not image_cfg["created"]:
        del image_cfg["created"]

    # finally write the image configuration
    imageconfig_hash, imageconfig_size = export2json(image_cfg, outdir)
    config = {
        "digest": f"sha256:{imageconfig_hash}",
        "mediaType": "application/vnd.oci.image.config.v1+json",
        "size": imageconfig_size,
    }

    # and write the manifest
    manifest = {
        "config": config,
        "layers": layers,
        "mediaType": "application/vnd.oci.image.manifest.v1+json",
        "schemaVersion": 2,
    }

    manif_hash, _ = export2json(manifest, outdir)

    return (manif_hash, imageconfig_hash)


def exportvm(
    vmname: str,
    outdir: str,
    libvirtURI: str | None = None,
    chunksize: int | None = None,
    vmconfig: dict = {},
    libvirtconfigfile: str | None = None,
    quiet: bool = False,
) -> str | None:
    """export a libvirt VM into an OCI compatible bundle.
    returns the hashdigest of the config
    or None.
    """
    # after a successful run we have:
    """
<outdir>/
<outdir>/oci-layout
<outdir>/index.json
<outdir>/blobs/sha256/<sha>   -> application/vnd.oci.image.manifest.v1+json
<outdir>/blobs/sha256/<sha:a> -> application/vnd.oci.image.config.v1+json
<outdir>/blobs/sha256/<sha:b> -> application/vnd.iem.sarde.config.v1+xml
<outdir>/blobs/sha256/<sha:x> -> application/application/x-qemu-disk+gzip
<outdir>/blobs/sha256/<sha:y> -> application/application/x-qemu-disk+gzip
<outdir>/blobs/sha256/<sha:z> -> application/application/gzip
[...]
    """
    # extract optional <tag> from container
    outcontainerdir, tag = utils.split_imagetag(outdir)
    if not issubclass(type(tag), str):
        tag = None

    oci_layout = {
        "imageLayoutVersion": "1.0.0",
    }
    container_index = {
        "schemaVersion": 2,
        "manifests": [],
    }

    if all([libvirtURI, libvirtconfigfile]):
        raise ValueError(
            f"libvirt connection {libvirtURI!r} and configuration {libvirtconfigfile!r} are mutually exclusive"
        )

    if not outcontainerdir:
        raise ValueError(f"output directory must not be empty: {outdir!r}")

    os.makedirs(outcontainerdir, exist_ok=True)

    (manifest_hash, imageconfig_hash) = (None, None)
    try:
        chunksdir = os.path.join(outcontainerdir, "blobs", "sha256")

        result = _exportVM(
            vmname,
            outdir=chunksdir,
            libvirtURI=libvirtURI,
            libvirtconfigfile=libvirtconfigfile,
            vmconfig=vmconfig,
            chunksize=chunksize,
            quiet=quiet,
        )
        try:
            manifest_hash, imageconfig_hash = result
        except TypeError:
            manifest_hash = imageconfig_hash = None
        if not imageconfig_hash:
            raise utils.TryException

        # read old oci-layout
        try:
            with open(os.path.join(outcontainerdir, "oci-layout"), "r") as f:
                oci_layout = json.load(f)
        except FileNotFoundError:
            pass
        except Exception as e:
            log.debug("Unable to read old oci-layout index", exc_info=True)

        # read old image manifest
        try:
            with open(os.path.join(outcontainerdir, "index.json"), "r") as f:
                container_index = json.load(f)
        except FileNotFoundError:
            pass
        except Exception as e:
            log.debug("Unable to read old container index", exc_info=True)

        manifest_stat = os.stat(os.path.join(chunksdir, manifest_hash))
        manifest = {
            "mediaType": "application/vnd.oci.image.manifest.v1+json",
            "digest": f"sha256:{manifest_hash}",
            "size": manifest_stat.st_size,
        }
        if tag:
            manifest["annotations"] = {
                "org.opencontainers.image.ref.name": tag,
            }

        container_index["manifests"].append(manifest)

        with open(os.path.join(outcontainerdir, "oci-layout"), "w") as f:
            json.dump(oci_layout, f, indent=2)

        with open(os.path.join(outcontainerdir, "index.json"), "w") as f:
            json.dump(container_index, f, indent=2)

    except utils.TryException:
        pass
    finally:
        utils.rmdirs(outcontainerdir, ["", "blobs", "sha256"])

    return imageconfig_hash


def _main():
    import argparse

    p = argparse.ArgumentParser(
        description="export a libvirt VM to an OCI-compliant directory.",
    )
    p.add_argument(
        "--connect",
        help="libvirt connection URI",
    )
    p.add_argument(
        "-o",
        "--outdir",
        default=".",
        help="output directory",
    )
    p.add_argument(
        "--chunksize",
        type=int,
        default=1024 * 1024,
        help="chunksize (in bytes)",
    )
    p.add_argument(
        "name",
        help="VM to export",
    )

    args = p.parse_args()

    x = exportvm(args.name, libvirtURI=args.connect, chunksize=args.chunksize)
    print(x)


if __name__ == "__main__":
    _main()



#!/usr/bin/env python3

# sardinecake - run GitLab-CI jobs within libvirt VMs
#
# Copyright © 2024, IOhannes m zmölnig, forum::für::umläute
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from logging import *
import logging

_log = None


class ColoredFormatter(logging.Formatter):
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"

    grey = "\033[38;20m"
    yellow = "\033[33;20m"
    green = "\033[92m"
    red = "\033[31;20m"
    bold_red = "\033[31;1m"
    dim = "\033[2m"
    reset = "\033[0m"

    def __init__(self, fmt=None, datefmt=None, style="%"):
        kwargs = {
            "style": style,
            "datefmt": datefmt or "%Y-%m-%d %H:%M:%S",
        }
        super().__init__(fmt=fmt, **kwargs)

        fmt = self._fmt

        self._formatter = {
            logging.DEBUG: logging.Formatter(self.dim + fmt + self.reset, **kwargs),
            # logging.INFO: logging.Formatter(self.green + fmt + self.reset, **kwargs),
            logging.INFO: logging.Formatter(self.reset + fmt + self.reset, **kwargs),
            logging.WARNING: logging.Formatter(
                self.yellow + fmt + self.reset, **kwargs
            ),
            logging.ERROR: logging.Formatter(self.red + fmt + self.reset, **kwargs),
            logging.CRITICAL: logging.Formatter(
                self.bold_red + fmt + self.reset, **kwargs
            ),
        }
        self._defaultformatter = logging.Formatter(**kwargs)

    def format(self, record):
        formatter = self._formatter.get(record.levelno, self._defaultformatter)
        return formatter.format(record)


def getLogger(name):
    global _log
    log = _log = logging.getLogger(name)
    logging.basicConfig()

    ch = logging.StreamHandler()
    ch.setFormatter(ColoredFormatter("%(asctime)s %(message)s"))

    master = logging.getLogger()
    for h in master.handlers:
        master.removeHandler(h)
    master.addHandler(ch)

    return log


def setVerbosity(verbosity: int | None, baselevel=logging.WARNING):
    if not verbosity:
        verbosity = 0
    if not _log:
        raise UserWarning("no logger initialized yet")
    _log.setLevel(max(1, baselevel - verbosity * 10))


if __name__ == "__main__":
    log = getLogger("test")
    setVerbosity(10)
    for name, level in logging.getLevelNamesMapping().items():
        log.log(level, name)



#!/usr/bin/env python3

# sardinecake - manage VMs via OCI
#
# Copyright © 2024, IOhannes m zmölnig, forum::für::umläute
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


# a generic domain representation (with backing as XML, libvirt,...)

import logging
import libvirt
from xml.dom import minidom
from xml.parsers import expat as expatparser

log = logging.getLogger("sardinecake.domain")


def _getChildrenByTagName(parent, tagName):
    """get all immediate children of <parent> that match the tagname"""
    for element in parent.getElementsByTagName(tagName):
        if element.parentNode == parent:
            yield element


def _addChild(parent, tagname):
    """add a new child Element with the specified tag to <parent>.
    returns the new child"""
    el = minidom.Element(tagname)
    parent.appendChild(el)
    el.ownerDocument = parent.ownerDocument
    return el


def _addBackingStore(parent, source, format):
    try:
        backingStore = parent.getElementsByTagName("backingStore")
        if backingStore:
            backingStore = backingStore[0]
        else:
            backingStore = _addChild(parent, "backingStore")

        for el in backingStore.childNodes:
            backingStore.removeChild(el)
        backingStore.setAttribute("type", "file")
        _addChild(backingStore, "format").setAttribute("type", format)
        _addChild(backingStore, "source").setAttribute("file", source)
    except:
        log.exception(f"adding backing file {backing_file} failed")
        backingStore = None
    return backingStore


def lookupVM(name, conn=None, quiet=False):
    if quiet:
        libvirt.registerErrorHandler((lambda ctx, err: 1), None)
    if not issubclass(type(conn), libvirt.virConnect):
        conn = libvirt.openReadOnly(conn)
    lookups = [conn.lookupByName, conn.lookupByUUIDString, conn.lookupByUUID]
    for lookup in lookups:
        try:
            return lookup(name)
        except libvirt.libvirtError:
            pass


class Domain:
    try:
        from typing import Self as _self

        _self = _self | None
    except ImportError:
        _self = "Domain"

    archmap = {
        # ###############################################
        # allowed values are (https://go.dev/doc/install/source#environment)
        # - "amd64" (64-bit x86, the most mature port)
        # - "386" (32-bit x86)
        # - "arm" (32-bit ARM)
        # - "arm64" (64-bit ARM)
        # - "ppc64le" (PowerPC 64-bit, little-endian)
        # - "ppc64" (PowerPC 64-bit, big-endian)
        # - "mips64le" (MIPS 64-bit, little-endian)
        # - "mips64" (MIPS 64-bit, big-endian)
        # - "mipsle" (MIPS 32-bit, little-endian)
        # - "mips" (MIPS 32-bit, big-endian)
        # - "s390x" (IBM System z 64-bit, big-endian)
        # - "wasm" (WebAssembly 32-bit)
        # ###############################################
        # keys found in /usr/share/libvirt/schemas/basictypes.rng
        "aarch64": "arm64",
        "alpha": "alpha",  # :-(
        "armv7l": "arm",
        "armv6l": "arm",
        "cris": "cris",  # :-(
        "i686": "386",
        "ia64": "ia64",  # :-(
        "lm32": "lm32",  # :-(
        "loongarch64": "loongarch64",  # :-(
        "m68k": "m68k",  # :-(
        "microblaze": "microblaze",  # :-(
        "microblazeel": "microblazeel",  # :-(
        "mips": "mips",
        "mipsel": "mipsle",
        "mips64": "mips64",
        "mips64el": "mips64le",
        "openrisc": "openrisc",  # :-(
        "parisc": "parisc",  # :-(
        "parisc64": "parisc64",  # :-(
        "ppc": "ppc",  # :-(
        "ppc64": "ppc64",
        "ppc64le": "ppc64le",
        "ppcemb": "ppcemb",  # :-(
        "riscv32": "riscv32",  # :-(
        "riscv64": "riscv64",  # :-(
        "s390": "s390",  # :-(
        "s390x": "s390x",
        "sh4": "sh4",  # :-(
        "sh4eb": "sh4eb",  # :-(
        "sparc": "sparc",  # :-(
        "sparc64": "sparc64",  # :-(
        "unicore32": "unicore32",  # :-(
        "x86_64": "amd64",
        "xtensa": "xtensa",  # :-(
        "xtensaeb": "xtensaeb",  # :-(
    }
    osmap = {
        # ###############################################
        # allowed values are (https://go.dev/doc/install/source#environment)
        # - "android"
        # - "darwin"
        # - "dragonfly"
        # - "freebsd"
        # - "illumos"
        # - "ios" ?
        # - "js"  ?
        # - "linux"
        # - "netbsd"
        # - "openbsd"
        # - "plan9" ?
        # - "solaris"
        # - "wasip1" ?
        # - "windows"
        # ###############################################
        # these are extracted from osinfo
        # check if '://<key>' is contained in the os ID
        # order matters!
        "android-x86.org": "android",
        "apple.com/": "darwin",
        "dragonflybsd.org/": "dragonfly",  # "dragonflybsd",
        "freebsd.org/": "freebsd",
        "freedos.org/": "dos",  # :-(
        "guix.gnu.org/guix/hurd": "hurd",  # :-(
        "guix.gnu.org/": "linux",
        "haiku-os.org/": "haiku",  # :-(
        "microsoft.com/msdos/": "dos",  # :-(
        "microsoft.com/win/1.": "win16",  # :-(
        "microsoft.com/win/2.": "win16",  # :-(
        "microsoft.com/win/3.": "win16",  # :-(
        "microsoft.com/win/9": "win9x",  # :-(
        "microsoft.com/win/me": "win9x",  # :-(
        "microsoft.com/": "windows",  # "winnt"
        "netbsd.org/": "netbsd",
        "novell.com/": "netware",  # :-(
        "omnios.org/": "illumos",
        "openbsd.org/": "openbsd",
        "openindiana.org/": "illumos",
        "oracle.com/solaris/": "solaris",
        "oracle.com/": "linux",
        "smartos.org/": "illumos",
        "sun.com/": "solaris",
        None: "linux",  # all the rest is linux
    }

    def __init__(self, xml: str):
        """create a Domain from a libvirt XML description"""
        self.xml = minidom.parseString(xml)

    def toXML(self):
        """get an XML representation (as a string) of the domain"""
        return self.xml.toxml()

    def __str__(self):
        """get a string representation (as XML)"""
        return self.toXML()

    def __repr__(self):
        return f"{type(self).__name__}({self.toXML()!r})"

    def getCPU(self):
        """get CPU family of the VM"""
        for os in self.xml.getElementsByTagName("os"):
            for typ in os.getElementsByTagName("type"):
                arch = typ.getAttribute("arch")
                if arch:
                    return self.archmap.get(arch, arch)

    def getOS(self):
        """get (lowercased) OS family of the VM"""
        for metadata in self.xml.getElementsByTagName("metadata"):
            for os in self.xml.getElementsByTagName("libosinfo:os"):
                osid = os.getAttribute("id")
                if osid:
                    for k, v in self.osmap.items():
                        if k and f"://{k}" in osid:
                            return v
        return self.osmap.get(None) or "unknown"

    def getDisks(self) -> dict[str, str]:
        """get all disks as dicts"""
        disks = []
        for diskType in self.xml.getElementsByTagName("disk"):
            disk = {}
            for diskNode in diskType.childNodes:
                name = diskNode.nodeName
                if name[0:1] == "#":
                    continue
                disk[name] = {
                    diskNode.attributes[attr].name: diskNode.attributes[attr].value
                    for attr in diskNode.attributes.keys()
                }
            disks.append(
                {
                    "type": diskType.getAttribute("type"),
                    "device": diskType.getAttribute("device"),
                    "file": disk.get("source", {}).get("file"),
                    "driver": disk.get("driver", {}).get("type"),
                    "target": disk.get("target", {}).get("dev"),
                }
            )
        return disks

    def getDiskFiles(self) -> dict[str, str]:
        """get the path to all file-based disks.
        returns a target-device -> source-file mapping
        """
        return {
            disk["target"]: disk["file"]
            for disk in self.getDisks()
            if disk.get("type") == "file" and disk.get("file")
        }

    def changeDiskFiles(
        self,
        disks: dict[str, str],
        backing_files: dict[str, str] = {},
        clear_old_backing_files: bool | None = True,
    ):
        """changes the disk-files according to the target-device -> source-file mapping in <disks>
        if <clear_old_backing_files> is True, remove all existing backing files
        if <clear_old_backing_files> is False, keep existing backing files
        if <backing_files> is None, keep existing (non-empty) backing files
        """
        changed = {}
        missed = {}
        if not backing_files:
            backing_files = {}
        for disk in self.xml.getElementsByTagName("disk"):
            if disk.getAttribute("type") != "file":
                continue
            target = None
            for t in disk.getElementsByTagName("target"):
                target = t.getAttribute("dev")
                if target in disks:
                    break
                target = None
            if not target:
                continue
            format = "qcow2"
            for d in disk.getElementsByTagName("driver"):
                format = d.getAttribute("type")
                break

            # remove empty backingStores
            if clear_old_backing_files is not False:
                for b in _getChildrenByTagName(disk, "backingStore"):
                    if not clear_old_backing_files:
                        if b.attributes.items():
                            continue
                        if b.hasChildNodes():
                            continue
                    log.debug(f"deleting backingStore node {b.toxml()!r}")
                    b.parentNode.removeChild(b)

            newsource = disks.get(target)
            backfiles = backing_files.get(target, [])
            if type(backfiles) == str:
                backfiles = [backfiles]
            missed[newsource] = True
            for s in disk.getElementsByTagName("source"):
                source = s.getAttribute("file")
                if source:
                    s.setAttribute("file", newsource)
                    changed[target] = newsource
                    missed[newsource] = False
                    # add backingstore if required
                    backparent = s.parentNode
                    for backing_file in backfiles:
                        backparent = _addBackingStore(backparent, backing_file, format)
                        if not backparent:
                            break

                    break

        # check if there were some elements that lacked a <source/> tag,
        # even though there was a matching <disk>
        disks = {k: v for k, v in disks.items() if missed.get(v)}
        if missed:
            log.debug(f"missed: {missed}")
            log.debug(f"missedisks: {disks}")
            for disk in self.xml.getElementsByTagName("disk"):
                if disk.getAttribute("type") != "file":
                    continue
                target = None
                for t in disk.getElementsByTagName("target"):
                    target = t.getAttribute("dev")
                    if target in disks:
                        break
                    target = None
                if not target:
                    continue
                newsource = disks.get(target)
                _addChild(disk, "source").setAttribute("file", newsource)
                changed[target] = newsource

        # return all the disks that have been changed
        return changed

    def getClonableDisks(self) -> dict:
        """check if all disks are either QCOW2-file-based disks or (implicitly sharable) CD-ROMs
        returns a dictionary of targetdev:qcow2image mappings
        exits if an unsupported device is found"""
        disks = {}
        for idx, d in enumerate(self.getDisks()):
            device = d.get("device")
            driver = d.get("driver")
            dtype = d.get("type")
            if device == "cdrom":
                continue
            if device != "disk":
                raise ValueError(f"Disk#{idx} is is an unsupported device {device!r}")
            if dtype != "file":
                raise ValueError(
                    f"Disk#{idx} is {dtype!r}-based (only 'file' is supported)"
                )
            # check if the disk is qcow2 based
            if driver != "qcow2":
                raise ValueError(
                    f"Disk#{idx} is of type {driver!r} (only 'qcow2' is supported)"
                )
            disks[d["target"]] = d["file"]
        return disks

    def getUUID(self) -> str | None:
        """get the UUID of the VM (if it has one)"""
        for uuidtag in _getChildrenByTagName(self.xml, "uuid"):
            return uuidtag.firstChild.wholeText

    def changeName(self, name: str | None) -> str | None:
        """change the <name> of the VM.
        returns the old name.
        if <name> is None, the name is not changed.
        on error, None is returned
        """
        nametag = None
        for nametag in self.xml.getElementsByTagName("name"):
            break
        if not nametag or not nametag.parentNode.parentNode == self.xml:
            return
        oldname = nametag.firstChild.wholeText
        if name is not None:
            nametag.firstChild.replaceWholeText(name)
        return oldname

    def getName(self) -> str | None:
        """get the <name> of the VM.
        on error, None is returned
        """
        return self.changeName(None)

    def anonymize(
        self, remove_uuid=True, remove_watchdog=True, mac_address: str | None = None
    ) -> bool:
        """remove instance specific information (like UUID, mac,...)
        use Domain.changeDiskFiles() and Domain.setName() to manually cleanup some more
        """
        domain = None
        for domain in self.xml.getElementsByTagName("domain"):
            break
        if not domain:
            return False

        if remove_uuid:
            for uuid in _getChildrenByTagName(domain, "uuid"):
                domain.removeChild(uuid)

        if remove_watchdog:
            # the 'itco' watchdog gives us trouble with older(?) libvirt
            for devices in _getChildrenByTagName(domain, "devices"):
                for watchdog in _getChildrenByTagName(devices, "watchdog"):
                    if watchdog.getAttribute("model") == "itco":
                        devices.removeChild(watchdog)

        for mac in domain.getElementsByTagName("mac"):
            if mac.parentNode.tagName != "interface":
                continue
            if mac.hasAttribute("address"):
                mac.removeAttribute("address")
            if mac_address:
                mac.setAttribute("address", mac_address)

        return True

    @classmethod
    def fromLibVirt(
        cls, conn: libvirt.virConnect | str | None, name: str, quiet=False
    ) -> _self:
        """create Domain from data fetched for libvirt VM <name>.
        if <conn> is a libvirt.virConnect, it is used (but not closed)
        otherwise a temporary readonly libvirt.virConnect is established (and closed) using <conn> as the connection specification.
        if <quiet> is True, libvirt errors are suppressed (effects future libvirt invocations)
        """
        dom = lookupVM(name=name, conn=conn, quiet=quiet)
        if dom:
            try:
                return cls(dom.XMLDesc())
            except expatparser.ExpatError:
                pass

    @classmethod
    def fromXMLFile(cls, filename: str) -> _self:
        """create Domain from XML file"""
        try:
            with open(filename) as f:
                xml = f.read()
        except OSError:
            return

        try:
            return cls(xml)
        except expatparser.ExpatError:
            pass


def _main():
    import argparse

    p = argparse.ArgumentParser(
        description="print some info about a libvirt Domain (VM)",
    )
    p.add_argument(
        "--connect",
        help="libvirt connection URI",
    )

    p.add_argument(
        "name",
        help="VM to export",
    )

    args = p.parse_args()

    with libvirt.openReadOnly(args.connect) as conn:
        d = Domain.fromLibVirt(conn, args.name)
    if not d:
        raise SystemExit(f"VM {args.name!r} does not exist")
    print(d)
    print("")
    print(f"CPU: {d.getCPU()}")
    print(f"OS : {d.getOS()}")
    print(f"disks: {d.getDiskFiles()}")
    print("")
    d.anonymize()
    print(d)


if __name__ == "__main__":
    _main()



#!/usr/bin/env python3

# gziparchive - poor mans archive with only gzip
#
# Copyright © 2024, IOhannes m zmölnig, forum::für::umläute
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# the gzip format is meant for compressing single files.
# it decidedly is **not** meant to compress multiple files and be able to
# reconstruct the individual files. this job is left for tar.
#
# nevertheless, this module tries to do exactly that:
# - compress multiple files (with no intermediate layer)
# - deflate the individual files
#
# it uses the fact that gzip streams have a header that (optionally) stores
# the original filename.
# also, gzip streams can be concatenated.
# so if we know the stream boundaries (or can guess them), we can restore the original
# files (but only the basenames)
#
# motivation: we want to be able to transmit difference files
# (for large disk images) efficiently.
#
# - our transport is OCI containers
# - to ease parallel downloads, a large file is split into multiple chunks
# - for QCOW2 we can use thin copies (which are effectively difference images)
#   to modify the large disk images without modifying the actual data.
# - for storing and downloading, we would like the original data to stay intact
#   (and reusable).
# - if we did a .tar.gz, this would modify the original data (at least the header(?))
# - since we are already chunking the data, we can conveniently start new chunks when
#   a new file is "added", thus helping the restore-process in guessing where the
#   next stream starts


import gzip
import hashlib
import logging
import os
import struct
import tempfile
import time


log = logging.getLogger("sardinecake.gziparchive")


def parseGZIPHeader(filename: str) -> dict:
    """parse the header of a filename to see if it is a GZip file
    returns a dict with the parsed information (or None)
    """
    OSs = {
        0: "FAT filesystem (MS-DOS, OS/2, NT/Win32)",
        1: "Amiga",
        2: "VMS (or OpenVMS)",
        3: "Unix",
        4: "VM/CMS",
        5: "Atari TOS",
        6: "HPFS filesystem (OS/2, NT)",
        7: "Macintosh",
        8: "Z-System",
        9: "CP/M",
        10: "TOPS-20",
        11: "NTFS filesystem (NT)",
        12: "QDOS",
        13: "Acorn RISCOS",
    }
    methods = {
        8: "DEFLATE",
    }

    FTEXT, FHCRC, FEXTRA, FNAME, FCOMMENT = 1, 2, 4, 8, 16

    def read_until_NULL(fp):
        data = None
        while True:
            s = fp.read(1)
            if not s or s == b"\000":
                break
            data = (data or b"") + s
        return data

    data = b""
    fextra = None
    fname = None
    fcomment = None
    fcrc = None
    with open(filename, "rb") as f:
        header = f.read(10)
        data += header

        try:
            (magic, method, flags, last_mtime, extra_flags, OS) = struct.unpack(
                "<HBBIBB", header
            )
        except struct.error:
            # couldn't read all data... not a GZip file
            return None
        if magic != 0x8B1F:
            # bag magic
            return None

        if flags & FEXTRA:
            # Read & discard the extra field, if present
            dat = f.read(2)
            data += dat
            (extra_len,) = struct.unpack("<H", dat)
            fextra = f.read(extra_len)
            data += fextra
            # TODO: parse extra data
        if flags & FNAME:
            # Read and discard a null-terminated string containing the filename
            fname = read_until_NULL(f)
            data += fname
        if flags & FCOMMENT:
            # Read and discard a null-terminated string containing a comment
            fcomment = read_until_NULL(f)
            data += fcomment
        if flags & FHCRC:
            # CRC16
            fcrc = f.read(2)
            data += fcrc

    result = {"magic": data[:2]}
    method = methods.get(method)
    if method:
        result["method"] = method
    if last_mtime:
        result["mtime"] = time.localtime(last_mtime)

    OS = OSs.get(OS)
    if OS:
        result["OS"] = OS

    if fextra:
        # TODO: parse extra data
        pass
    if fname:
        result["name"] = fname.decode(encoding="iso-8859-1")
    if fcomment:
        result["comment"] = fcomment.decode(encoding="iso-8859-1")
    if fcrc:
        # TODO: check CRC
        pass

    return result


def parseGZIPFooter(filename: str) -> dict:
    """parse the footer (END) of a filename assuming it is a GZip file.
    returns a dict with the parsed information (or None)
    """
    with open(filename, "rb") as f:
        try:
            f.seek(0, os.SEEK_END)
            f.seek(f.tell() - 8, os.SEEK_SET)
        except ValueErrir:
            return None
        data = f.read(8)
        return {
            "crc": data[:4],
            "size": struct.unpack("<I", data[4:]),  # size of the uncompressed data!
        }


class FileBucketReader:
    def __init__(self, indir, filenames):
        self._indir = indir
        self._filenames = iter(filenames)
        self._fd = None
        self._opened = []
        self._opened += self._next()

    def _next(self):
        if self._fd:
            self._fd.close()
        self._fd = None
        try:
            filename = next(self._filenames)
        except StopIteration:
            return False
        self._fd = open(os.path.join(self._indir, filename), "rb")
        return filename

    def read(self, size=-1):
        if not self._fd:
            if not self._next():
                # EOFs
                return b""

        if size > 0:
            ret = self._fd.read(size)
            if not ret:
                # no data read, maybe we just hit the EOF of the current file
                if not self._next():
                    return ret
                ret = self._fd.read(size)
        else:
            #  wants to read *all* data?
            ret = b""
            while self._fd:
                r = self._fd.read(size)
                if r:
                    ret += r
                else:
                    self._next()
        return ret

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        if self._fd:
            self._fd.close()
        self._fd = None


class FileBucketWriter:
    def __init__(self, outdir, chunksize=None):
        if chunksize and chunksize < 0:
            raise ValueError(f"chunksize must be positive or 0 (got {chunksize})")

        os.makedirs(outdir, exist_ok=True)
        self._outdir = outdir
        self._chunksize = chunksize

        self._hashes = []
        self._currentsize = 0

        self._file = None
        self._hasher = None

    def _write(self, buffer):
        bufsize = len(buffer)
        self._hasher.update(buffer)
        self._file.write(buffer)
        self._currentsize += bufsize

    def _close(self):
        if self._hasher:
            hashstr = self._hasher.hexdigest()
            self._hashes.append(hashstr)
            if self._file:
                try:
                    os.link(self._file.name, os.path.join(self._outdir, hashstr))
                    log.info(f"wrote chunk #{len(self._hashes):03}: {hashstr!r}")
                except FileExistsError:
                    # file exists; since the filename is a hash, this means that the it is the same file
                    log.info(
                        f"skipped existing chunk #{len(self._hashes):03}: {hashstr!r}"
                    )
                    pass
        self._hasher = None
        self._file = None

    def _next(self):
        """finish up the current file, open the next one"""
        self._close()

        os.makedirs(self._outdir, exist_ok=True)
        self._hasher = hashlib.new("SHA256")
        self._file = tempfile.NamedTemporaryFile(dir=self._outdir, prefix=".")
        self._currentsize = 0

    def write(self, buffer):
        """write some data"""
        bufsize = len(buffer)
        # if chunksize:
        #    sizefree = self._chunksize - self._currentsize
        #    if size < sizefree:

        if not self._file:
            self._next()

        if self._chunksize:
            while buffer:
                size = len(buffer)
                maxsize = min(size, self._chunksize - self._currentsize)

                if maxsize <= 0:
                    # currentsize => chunksize: proceed to next file
                    self._next()
                    continue

                buf = buffer[:maxsize]
                buffer = buffer[maxsize:]
                self._write(buf)
        else:
            # not chunking; just write the data to the file
            self._write(buffer)

        return bufsize

    def close(self):
        """close the remaining file-handle, return a list of created hashes"""
        self._close()
        return self._hashes

    def hashes(self):
        """return a list of created hashes so far"""
        return self._hashes

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.close()


def gzipchunked(
    infilename,
    outdirname,
    compresslevel=6,
    mtime=None,
    chunksize: int | None = None,
    name: str | None = None,
):
    """compressed the <infilename> with gzip, writing the result to <outdirname>/
    if the <outdirname> directory does not exist it will be created.
    the resulting compressed files will have a maximum size of <chunksize>,
    multiple files are created if needed.
    if chunksize is falsish, no re-chunking is performed.

    the filenames of the split gzip compressed data, is the SHA256 hexdigest of the compressed data.

    if <mtime> is falsish, it is read from <infilename> (otherwise it behaves the same as with gzip.GzipFile)
    compresslevel behaves the same as gzip.GzipFile.
    if <name> is not None, it is used for embedding the filename into the gzip stream.
    if <name> is None, <infilename> is used instead

    returns the names (hashes, without the directory) of the files written.
    """
    import shutil

    if not mtime:
        mtime = os.stat(infilename).st_mtime
    log.debug(f"gzipchunk(infile={infilename!r}, outdir={outdirname!r}, name={name!r})")
    if name is None:
        name = infilename
    count = 0
    with (
        open(infilename, "rb") as f_in,
        FileBucketWriter(outdirname, chunksize=chunksize) as f_chunky,
        gzip.GzipFile(
            filename=name,
            fileobj=f_chunky,
            mode="wb",
            compresslevel=compresslevel,
            mtime=mtime,
        ) as f_out,
    ):
        shutil.copyfileobj(f_in, f_out)
        hashes = f_chunky.hashes()
    return hashes


def gunzipchunked(indirname, outdirname, bucketfiles):
    """decompress the <bucketfiles> in <indirname> into files in <outdirname>

    the actual compressed files can be split into multiple <bucketfiles>
    <bucketfiles> must have the correct order.

    <bucketfiles> might contain multiple compressed files.
    """
    import shutil

    # partition the bucketfiles to start with a gzip header
    gzipheaders = [
        parseGZIPHeader(os.path.join(indirname, filename)) for filename in bucketfiles
    ]
    headerfiles = [
        filename for filename, headers in zip(bucketfiles, gzipheaders) if headers
    ]

    buckets = []
    metadata = []
    bucket = []
    for b, h in zip(bucketfiles, gzipheaders):
        if h:
            # start a new bucket
            bucket = []
            buckets.append(bucket)
            metadata.append(h)
        bucket.append(b)

    os.makedirs(outdirname, exist_ok=True)

    unzippedfiles = []
    for b, h in zip(buckets, metadata):
        mtime = h.get("mtime")
        name = h.get("name")
        if name:
            name = os.path.basename(name)
            outfilename = os.path.join(outdirname, name)
            f_out = open(outfilename, "wb")
        else:
            f_out = tempfile.NamedTemporaryFile(dir=outdirname, delete=False)
            outfilename = f_out.name
        with (
            FileBucketReader(indirname, b) as f_chunky,
            gzip.GzipFile(fileobj=f_chunky, mode="rb") as f_in,
        ):
            shutil.copyfileobj(f_in, f_out)
        f_out.close()
        if mtime and outfilename:
            mtime = time.mktime(mtime)
            os.utime(outfilename, times=(mtime, mtime))
        unzippedfiles.append(outfilename)

    return unzippedfiles


def _tests():
    _testdata = bytes(range(256)) * 4
    testdata = {"count": 0}

    def getOutdir():
        testcount = testdata["count"] = testdata.get("count", 0) + 1
        return f"/tmp/gzipchunk/{testcount}"

    def _test1(outpath, chunksize=None):
        outfile = FileBucketWriter(outpath, chunksize=chunksize)
        with gzip.GzipFile(filename="blabla", fileobj=outfile, mode="wb") as gz:
            gz.write(_testdata)
        hashes = outfile.close()
        print(" ".join(hashes))

    def _test2(outpath, chunksize=None):
        with (
            FileBucketWriter(outpath, chunksize=chunksize) as outfile,
            gzip.GzipFile(filename="blabla", fileobj=outfile, mode="wb") as gz,
        ):
            gz.write(_testdata)
            hashes = outfile.hashes()
        print(" ".join(hashes))

    def _test3(infilename, outpath, chunksize=None):
        with (
            open(infilename, "rb") as infile,
            FileBucketWriter(outpath, chunksize=chunksize) as outfile,
            gzip.GzipFile(filename=infilename, fileobj=outfile, mode="wb") as gz,
        ):
            gz.write(infile.read())
            hashes = outfile.hashes()
        print(" ".join(hashes))

    def _test4(infilename, outpath, chunksize=None):
        import shutil

        with (
            open(infilename, "rb") as infile,
            FileBucketWriter(outpath, chunksize=chunksize) as outfile,
            gzip.GzipFile(filename=infilename, fileobj=outfile, mode="wb") as gz,
        ):
            shutil.copyfileobj(infile, gz)
            hashes = outfile.hashes()
        print(" ".join(hashes))

    def _test5(infilename, outpath, chunksize=None):
        import shutil

        mtime = os.stat(infilename).st_mtime
        with (
            open(infilename, "rb") as infile,
            FileBucketWriter(outpath, chunksize=chunksize) as outfile,
            gzip.GzipFile(
                filename=infilename, fileobj=outfile, mode="wb", mtime=mtime
            ) as gz,
        ):
            shutil.copyfileobj(infile, gz)
            hashes = outfile.hashes()
        print(" ".join(hashes))

    def _test6(infilename, outpath, chunksize=None):
        hashes = gzipchunked(infilename, outpath, chunksize=chunksize)
        print(" ".join(hashes))

    _test1(getOutdir())
    _test1(getOutdir(), 50)
    _test2(getOutdir())
    _test2(getOutdir(), 50)
    _test3(__file__, getOutdir())
    _test3(__file__, getOutdir(), 512)
    _test4(__file__, getOutdir())
    _test4(__file__, getOutdir(), 512)
    _test5(__file__, getOutdir())
    _test5(__file__, getOutdir(), 512)
    _test6(__file__, getOutdir())
    _test6(__file__, getOutdir(), 512)


if __name__ == "__main__":
    _tests()



#!/usr/bin/env python3

# sardinecake - run GitLab-CI jobs within libvirt VMs
#
# Copyright © 2024, IOhannes m zmölnig, forum::für::umläute
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess

from contextlib import contextmanager
import logging as logging

log = logging.getLogger("sardinecake.ssh")


@contextmanager
def SSH(*args, **kwargs):
    import paramiko

    class IgnorePolicy(paramiko.client.MissingHostKeyPolicy):
        """
        Policy that silently ignores any missing host keys.
        """

        def missing_host_key(self, client, hostname, key):
            log.debug(
                "Unknown {} host key for {}: {}".format(
                    key.get_name(),
                    hostname,
                    paramiko.client.hexlify(key.get_fingerprint()),
                )
            )

    if "port" in kwargs and not kwargs["port"]:
        del kwargs["port"]

    try:
        client = paramiko.SSHClient()
        client.load_system_host_keys()
        client.set_missing_host_key_policy(IgnorePolicy)
        client.connect(*args, **kwargs)
        yield client
    finally:
        client.close()


def checkSSH(
    host: str,
    username: str | None = None,
    password: str | None = None,
    key_filename: str | list[str] | None = None,
    timeout: float | None = None,
    port: int | None = None,
) -> bool | None:
    """test whether we can open an SSH connection to <host>"""
    try:
        with SSH(
            host,
            username=username,
            password=password,
            key_filename=key_filename,
            timeout=timeout,
            port=port,
        ) as client:
            pass
    except TimeoutError as e:
        log.error("SSH connection to VM timed out!")
        log.debug(f"ssh://{username}@{host} timed out with {e}")
        return False
    except Exception as e:
        log.exception(f"ssh://{username}@{host} failed with {e}")
        return None
    return True


def channel2file(inchan, outfile):
    """write data received from <inchan> to <outfile>"""
    written = 0
    while inchan.recv_ready():
        written += outfile.write(inchan.recv(65536))
    return written


def getbashexports(env={}):
    if not env:
        return b""
    cmd = subprocess.run(
        ["bash", "-c", "set"],
        stdout=subprocess.PIPE,
        env={str(k): str(v) for k, v in env.items()},
    )
    result = []
    for line in cmd.stdout.splitlines():
        k = line.split(b"=", maxsplit=1)
        if k[0].decode() in env:
            result.append(b"export " + line)
    return b"\n".join(result)


def run_commands(host, commands, **kwargs):
    """execute (on <host>) the given <commands> within a bash
    <kwargs> are passed to SSH.
    """

    ret = None
    with (
        open("/dev/stdout", "wb", buffering=0) as binout,
        SSH(host, **kwargs) as client,
    ):
        stdin, stdout, stderr = client.exec_command("/bin/bash", get_pty=False)
        stdin.write(commands)
        while not stdout.channel.exit_status_ready():
            channel2file(stdout.channel, binout)
        channel2file(stdout.channel, binout)
        ret = stdout.channel.exit_status

    return ret


def run_script(host, script, env={}, **kwargs):
    """copy script to <host> and execute it; possibly sourcing a file containing the <env> vars first"""
    dirname = f"gitlab-{os.getpid()}"
    scriptname = os.path.join(dirname, "script")
    envname = os.path.join(dirname, "wrapper")

    log.debug(f"run {script} on {host}")
    ret = None
    with (
        open(script) as f,
        open("/dev/stdout", "wb", buffering=0) as binout,
        SSH(host, **kwargs) as client,
    ):
        sftp = client.open_sftp()
        # (try to) touch /var/log/lastlog, so ssh does not not complain if it is missing
        try:
            with sftp.open("/var/log/lastlog", "ab") as lastlog:
                pass
        except OSError as e:
            log.debug(f"couldn't touch lastlog: {e}")

        try:
            sftp.mkdir(dirname)
            log.debug(f"creating script {scriptname!r}")
            sftp.putfo(f, scriptname)
            sftp.chmod(scriptname, 0o711)
            log.debug(f"creating env-file {envname!r}")
            with sftp.open(envname, "wb") as envfd:
                envfd.write(b"#!/usr/bin/env bash\n")
                envfd.write(
                    b"if set -o | grep pipefail > /dev/null; then set -o pipefail; fi\n"
                )
                envfd.write(b"set -o errexit\n")
                envfd.write(b"set +o noclobber\n\n")
                envfd.write(getbashexports(env))
                envfd.write(b"\n\n")
                if False:
                    # this gobbles the script's exit-code
                    envfd.write(
                        b""": | eval $'%s\\n' 2>&1\nexit 0\n""" % scriptname.encode()
                    )
                else:
                    # this preserves the exit code
                    envfd.write(b"set +o errexit\n")
                    envfd.write(b""": | eval $'%s\\n' 2>&1\n""" % scriptname.encode())

            sftp.chmod(envname, 0o711)

            stdin, stdout, stderr = client.exec_command(
                envname, get_pty=False, environment=env
            )
            while not stdout.channel.exit_status_ready():
                channel2file(stdout.channel, binout)
            channel2file(stdout.channel, binout)
            ret = stdout.channel.exit_status
        finally:
            try:
                sftp.unlink(scriptname)
                sftp.unlink(envname)
                sftp.rmdir(dirname)
            except Exception as e:
                log.debug(f"Failed to cleanup {dirname!r} on {host!r}: {e}")

    return ret


if __name__ == "__main__":
    import argparse
    import sys
    from urllib.parse import urlparse

    logging.basicConfig()

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--timeout",
        type=float,
        help="timeout for SSH-connection (DEFAULT: no timeout)",
    )
    p = parser.add_argument_group("script")
    p.add_argument(
        "--script",
        help="script to run on the server",
    )
    p.add_argument(
        "--env",
        type=lambda x: x.split("=", maxsplit=1),
        default=[],
        action="append",
        help="environment for the script (use 'VAR=value'); can be given multiple times",
    )
    parser.add_argument(
        "URI",
        type=urlparse,
        help="SSH-URI to check for connectivity (e.g. 'ssh://user:password@example.com')",
    )

    p = parser.add_argument_group("verbosity")
    p.add_argument(
        "-v",
        "--verbose",
        action="count",
        help="Raise verbosity (can be given multiple times)",
    )
    p.add_argument(
        "-q",
        "--quiet",
        action="count",
        help="Lower verbosity (can be given multiple times)",
    )

    args = parser.parse_args()

    # merge verbosity settings
    if args.verbose or args.quiet:
        args.verbosity = (args.verbose or 0) - (args.quiet or 0)
        verbosity = args.verbosity
    else:
        args.verbosity = None
        verbosity = 0  # defaults.verbosity
    # logging.setVerbosity(verbosity)
    log.setLevel(max(1, logging.INFO - 10 * verbosity))
    del args.quiet
    del args.verbose

    uri = args.URI
    if uri.scheme != "ssh":
        parser.exit(1, "URI-schema must be 'ssh'")

    env = {}
    for kv in args.env:
        try:
            k, v = kv
        except ValueError:
            log.error(f"env must be a 'key=value' pair... ignoring {kv[0]!r}")
            continue
        env[k] = v

    if args.script:
        try:
            with open(args.script) as f:
                pass
        except Exception as e:
            parser.exit(1, f"build-script cannot be opened: {e}")

    if checkSSH(
        uri.hostname, username=uri.username, password=uri.password, port=uri.port
    ):
        print(f"SSH connection to {uri.hostname!r} OK.", file=sys.stderr)
    else:
        raise SystemExit(f"couldn't establish an SSH connection to {uri.hostname}.\n")

    if args.script:
        ret = run_script(
            uri.hostname,
            args.script,
            env=env,
            username=uri.username,
            password=uri.password,
            port=uri.port,
        )
        if ret:
            parser.exit(ret, f"script exited with: {ret}\n")



#!/usr/bin/env python3

# sardinecake - manage VMs via OCI
#
# Copyright © 2024, IOhannes m zmölnig, forum::für::umläute
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


# push/pull interaction with an OCI registry

# skopeo-copy(1)
# container-transports(5)
# `skopeo copy "docker://alpine:${tag}" "oci:OCI/alpine:${tag}"`
# this fetches the raw blobs to ./OCI/alpine (the OCI dir must exist!))
# - multiple tags can be copied to the same directory (index.json then contains multiple manifests)

import os
import logging
import subprocess

try:
    from . import utils
except ImportError:
    import utils


log = logging.getLogger("sardinecake.ociregistry")


def fetch(container: str, ocidir: str = "", sourcecontainer: str | None = None) -> bool:
    """fetch a container from a registry"""
    srccontainer = sourcecontainer or container
    contain, tag = utils.split_imagetag(container)
    if not issubclass(type(tag), str):
        container = contain
    dstcontainer = os.path.join(ocidir or "", container)
    dstcontainerdir = os.path.join(ocidir or "", contain)

    # skopeo copy docker://ghcr.io/cirruslabs/macos-monterey-vanilla:latest oci:monterey:latest
    os.makedirs(dstcontainerdir, exist_ok=True)
    cmd = ["skopeo", "copy", f"docker://{srccontainer}", f"oci:{dstcontainer}"]
    log.debug(cmd)

    args = {}
    if log.getEffectiveLevel() >= logging.INFO:
        args["stdout"] = subprocess.DEVNULL
        args["stderr"] = subprocess.DEVNULL

    p = subprocess.run(cmd, **args)
    if p.returncode:
        raise SystemExit(p.returncode)
    return True


def push(container: str, ocidir: str = "", targetcontainer: str | None = None) -> bool:
    """push a container to a registry"""
    srccontainer = os.path.join(ocidir or "", container)
    dstcontainer = targetcontainer or container

    cmd = ["skopeo", "copy"]
    cmd += [f"oci:{srccontainer}", f"docker://{dstcontainer}"]
    log.debug(cmd)

    p = subprocess.run(cmd)
    if p.returncode:
        raise SystemExit(p.returncode)
    return True


def delete(container: str, ocidir: str | None = None) -> bool:
    """delete a container from the registry (if ocidir is None) or the local directory"""
    if ocidir:
        container = f"oci:{os.path.join(ocidir, container)}"
    else:
        container = f"docker://{container}"

    cmd = ["skopeo", "delete", container]
    p = subprocess.run(cmd)
    if p.returncode:
        raise SystemExit(p.returncode)
    return True


def _main():
    import argparse

    parser = argparse.ArgumentParser(
        description="fetch/push a sardine container from/to a registry.",
    )

    p = parser.add_argument_group("action")
    p = p.add_mutually_exclusive_group()
    p.add_argument(
        "--fetch",
        action="store_true",
        default=True,
        help="fetch container from registry (DEFAULT)",
    )
    p.add_argument(
        "--push",
        action="store_false",
        dest="fetch",
        help="push container to registry",
    )

    parser.add_argument(
        "container",
        help="container to push/fetch",
    )

    args = parser.parse_args()

    if args.fetch:
        x = fetch(args.container)
    else:
        x = push(args.container)
    print(x)


if __name__ == "__main__":
    _main()



#!/usr/bin/env python3

# sardinecake - manage VMs via OCI
#
# Copyright © 2024, IOhannes m zmölnig, forum::für::umläute
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

_version = None

import json
import logging
import os
import pathlib

from .utils import parse_size, checkSchema
from . import defaults

try:
    from ._version import version as _version
except ImportError:
    pass

log = logging.getLogger("sardinecake")
logging.basicConfig()

SYSTEM_FAILURE = 1

example_container = "registry.example.org/sardines/debian:stable"


def do_selftest(args):
    """perform some tests (mostly about logging) and printout stuff"""
    import pprint
    import sys

    glob = dict(globals())
    del glob["__builtins__"]

    log.error(f"NAME: {__name__!r}")
    log.error(f"FILE: {__file__!r}")

    log.warning(pprint.pformat(glob, sort_dicts=False))
    log.warning(pprint.pformat(sys.argv, sort_dicts=False))

    for name, level in logging.getLevelNamesMapping().items():
        log.log(level, name)

    print(f"log-level: {log.level}/{log.getEffectiveLevel()}")
    print(args)


def do_login(args):
    from .auth import login

    registry = args.registry
    username = args.username
    password = args.password
    log.debug(f"LOGIN to {registry!r} as {username!r}")
    if not login(registry, username=username, password=password):
        raise SystemExit(f"Failed to login to {registry!r}")


def do_logout(args):
    from .auth import logout

    registry = args.registry
    if registry == "*":
        registry = None
    log.debug(f"LOGOUT from {registry!r}")
    logout(registry)


def do_fetch(args):
    from .ociregistry import fetch

    container = args.container
    ocidir = args.ocidir
    sourcecontainer = args.source
    log.debug(f"FETCH {sourcecontainer!r} as {container!r} into {ocidir!r}")
    if not fetch(container, ocidir=ocidir, sourcecontainer=sourcecontainer):
        raise SystemExit(f"failed to fetch {container!r}")


def do_push(args):
    from .ociregistry import push

    container = args.container
    ocidir = args.ocidir
    targetcontainer = args.target
    log.debug(f"PUSH {container!r} from {ocidir!r} to {targetcontainer!r}")
    if not push(container, ocidir=ocidir, targetcontainer=targetcontainer):
        raise SystemExit(f"failed to push {container!r}")


def do_list(args):
    # list all (locally) available images
    # --------------------------
    # docker: | REPOSITORY | TAG | IMAGE ID  (short config digest) | CREATED | SIZE |
    # docker --digest:  | REPOSITORY | TAG | (manifest) DIGEST | IMAGE ID (short config digest) | CREATED | SIZE |
    #
    # skopeo get by hash: container@sha256:<1234567890>; where <1234567890> is the *full* manifest digest
    # docker get by hash: container@sha256:<1234567890>; where <1234567890> is the *full* manifest digest
    # we can run:
    # - docker run <container>:<tag>
    # - docker run <container>@sha256:<manifestdigest>
    # - docker run <shortconfigdigest>
    # - docker run <any length of the config digest that is unique)
    # gitlab-docker-executor prints:
    # "Using docker image sha256:<configdigest> for <name>:<tag> with digest <name>@sha256:<manifestdigest> ..."
    #
    # --------------------------
    # tart  : | source | name | disk | size | state |
    #
    # the <name> lists all tags (in the form <container>:<tag>) and the corresponding hash
    # (in the form <container>@sha256:<manifesthash>)
    # --------------------------
    # sarde : ?
    # for the text form, we probably should go with 'tart' (but leaving out disk resp size; or put it at the end)
    # | source | name | state |
    # the JSON form should contain the name and (both) the hashes separately
    from .importvm import getconfig
    from .domain import lookupVM

    log.debug(f"LIST containers in {args.ocidir!r}")
    title = {
        "source": "SOURCE",
        "name": "NAME",
        "id": "IMAGE ID",
        "state": "STATE",
        "tag": "",
        "digest": "",
    }
    states = {
        True: "running",
        False: "stopped",
        None: "missing",
    }

    root = pathlib.Path(args.ocidir)
    dirs = [
        p.parent
        for p in root.glob("**/oci-layout")
        if (p.parent / "index.json").exists()
    ]
    # (container, manifest-file, tag)
    sardines = []
    for p in dirs:
        with (p / "index.json").open() as f:
            name = p.relative_to(root)
            index = json.load(f)

        for m in index["manifests"]:
            tag = m.get("annotations", {}).get("org.opencontainers.image.ref.name")
            digest_m = m.get("digest")
            manifestpath = p / "blobs" / os.path.join(*digest_m.split(":"))
            with manifestpath.open() as f:
                manifest = json.load(f)
            digest_c = manifest["config"]["digest"]
            state = None
            dom = lookupVM(name=digest_c, conn=args.connect, quiet=True)
            if dom:
                state = bool(dom.isActive())
            sardines.append(
                {
                    "source": "oci",
                    "name": str(name),
                    "tag": tag,
                    "digest": digest_m,
                    "id": digest_c,
                    "state": state,
                }
            )

    if args.format == "json":
        print(json.dumps(sardines, indent=2))
    else:
        fieldsize = {k: len(v) for k, v in title.items()}
        for s in sardines:
            for k, v in s.items():
                fieldsize[k] = max(fieldsize.get(k, 0), len(str(v)))
                fieldsize[k] = max(fieldsize.get(k, 0), len(str(k)))

        formatstr = "%s\t%s\t%s"
        size_source = fieldsize["source"]
        size_name = max(fieldsize["tag"], fieldsize["digest"]) + fieldsize["name"] + 1
        size_id = fieldsize["id"]
        size_state = fieldsize["state"]
        formatstr = f"%-{size_source}s\t%-{size_name}s\t%-{size_id}s\t%-{size_state}s"

        sardines = sorted(sardines, key=lambda x: (x["id"], x["name"], x["tag"] or ""))
        lastid = None
        print(formatstr % (title["source"], title["name"], title["id"], title["state"]))
        for sardine in sardines:
            if lastid != sardine["id"]:
                lastid = sardine["id"]
                name = f"{sardine['name']}@{sardine['digest']}"
                state = states.get(sardine["state"], states[None])
                print(formatstr % (sardine["source"], name, sardine["id"], state))
            if sardine["tag"]:
                name = f"{sardine['name']}:{sardine['tag']}"
                state = '   "   '
                print(formatstr % (sardine["source"], name, sardine["id"], state))


def do_getconfig(args):
    from .importvm import getconfig

    if args.raw:
        ocidir = args.container
    else:
        ocidir = os.path.join(args.ocidir or "", args.container)

    log.debug(f"GETCONFIG for {args.container!r} in {args.ocidir!r}")
    result = getconfig(
        ocidir=ocidir,
    )
    if result is None:
        log.debug(f"no configuration found for {args.container!r}")
        return
    print(json.dumps(result, indent=2))


def do_clone(args):
    from .importvm import importvm, getconfig
    from . import clone

    log.debug(f"CLONE {args}")

    # first make sure that the source container is available
    try:
        raw = args.raw
    except:
        raw = False

    if raw:
        ocidir = None
    else:
        ocidir = args.ocidir or ""

    container = args.container
    vmdir = args.vmdir

    try:
        cfg = clone.fetch_and_clone(
            container=container,
            vm=args.target,
            ocidir=ocidir,
            vmdir=vmdir,
            libvirtURI=args.connect,
        )
    except Exception as e:
        log.debug("clone failed", exc_info=True)
        raise SystemExit(e)
    print(json.dumps(cfg, indent=2))


def do_export(args):
    from .exportvm import exportvm

    name = args.name
    container = args.container or name
    libvirtURI = args.connect
    libvirtConfig = args.libvirtconfig
    if libvirtConfig:
        libvirtURI = None

    log.debug(f"EXPORT {name!r} to {args.ocidir!r} as {container!r}")

    try:
        raw = args.raw
    except:
        raw = False

    vmconfig = None
    if args.vmconfig:
        with open(args.vmconfig) as f:
            vmconfig = json.load(f)
            check = checkSchema(vmconfig, "sardineconfig")
            if check is False:
                log.fatal(
                    f"VM-config {args.vmconfig!r} does not comply with the sardine-config schema"
                )
                raise SystemExit(1)
            if not check:
                log.warning(
                    f"Could not verify VM-config {args.vmconfig!r} against schema"
                )

    if raw:
        ocidir = container
    else:
        ocidir = os.path.join(args.ocidir or "", container)
    chunksize = args.chunksize

    log.debug(f"exporting {name!r} to {ocidir!r} with chunks of {chunksize} bytes")

    result = exportvm(
        name,
        outdir=ocidir,
        libvirtURI=libvirtURI,
        libvirtconfigfile=libvirtConfig,
        chunksize=chunksize,
        vmconfig=vmconfig,
        quiet=log.getEffectiveLevel() >= logging.INFO,
    )
    if result is None:
        raise SystemExit(1)
    print(result)


def do_import(args):
    from .importvm import importvm

    try:
        raw = args.raw
    except:
        raw = False

    if raw:
        ocidir = args.container
    else:
        ocidir = os.path.join(args.ocidir or "", args.container)

    vmdir = args.vmdir
    name = args.name

    log.debug(f"IMPORT {args.container!r} from {args.ocidir} to {vmdir!r} as {name!r}")

    result = importvm(
        ocidir=ocidir,
        outdir=vmdir,
        name=name,
        libvirtURI=args.connect,
        quiet=log.getEffectiveLevel() >= logging.INFO,
    )
    if result is None:
        raise SystemExit(1)
    try:
        if args.print_uuid:
            print(result[1])
        else:
            print(result[0])
    except IndexError:
        print(result)


def do_pull(args):
    log.debug(f"PULL {args}")

    do_fetch(args)
    do_import(args)


def readConfig(configfiles: list[str], with_default=False) -> dict:
    import configparser

    converters = {
        "verbose": int,
        "quiet": int,
        "chunksize": parse_size,
        "vmdir": os.path.expanduser,
        "ocidir": os.path.expanduser,
    }

    config = configparser.ConfigParser()
    config.read(os.path.expanduser(f) for f in configfiles)

    configuration = {
        section: dict(config[section].items())
        for section in config
        if with_default or config.default_section != section
    }

    defaults = configuration.get(config.default_section)
    if defaults:
        del configuration[config.default_section]
    configuration[None] = defaults

    # fix types
    for name, section in configuration.items():
        for k, converter in converters.items():
            if not converter:
                continue
            if k in section:
                section[k] = converter(section[k])

    return configuration


def parseArgs(default):
    import argparse

    formatter = argparse.RawDescriptionHelpFormatter

    def add_common_args(parser: argparse.ArgumentParser):
        p = parser.add_argument_group("libvirt")
        p.add_argument(
            "--connect",
            help="libvirt connection URI",
        )

        p = parser.add_argument_group("verbosity")
        p.add_argument(
            "-v",
            "--verbose",
            action="count",
            help="Raise verbosity (can be given multiple times)",
        )
        p.add_argument(
            "-q",
            "--quiet",
            action="count",
            help="Lower verbosity (can be given multiple times)",
        )
        p.add_argument(
            "--no-run",
            default=None,
            action="store_true",
            help="Don't actually run (exit early)",
        )

        return parser

    # common args for all sub-parsers and the top-level parser
    common = argparse.ArgumentParser(add_help=False)
    add_common_args(common)

    parser = argparse.ArgumentParser(
        description="manage libvirt VMs via OCI registries",
        parents=[common],
    )

    def add_default_args(subparsers, name=None, fallbacks={}, func=None, **kwargs):
        if any([name, func]) and not all([subparsers, name, func]):
            raise ValueError("inconsistent arguments for add_subparser()")

        if name:
            if "help" in kwargs and "description" not in kwargs:
                kwargs["description"] = kwargs["help"]
            p = subparsers.add_parser(name, **kwargs)
            add_common_args(p)
        else:
            p = subparsers

        fallbacks.setdefault("ocidir", defaults.sarde.oci_path)
        fallbacks.setdefault("vmdir", defaults.sarde.vmdisk_path)
        fallbacks.setdefault("verbose", 0)
        fallbacks.setdefault("quiet", 0)
        mydefaults = default.get(name) or default.get(None) or {}
        fallbacks.update(mydefaults)
        fallbacks["func"] = func

        if func:
            p.set_defaults(**fallbacks)

        return p

    add_default_args(parser)

    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=_version or "<unknown>",
    )

    subparsers = parser.add_subparsers(required=True, title="commands")

    # ################### login ###################
    p = add_default_args(
        subparsers,
        "login",
        func=do_login,
        help="Login to a registry",
        formatter_class=formatter,
        epilog="""
For pushing images to a registry (and sometimes for pulling),
you often need to authenticate yourself. Do this via the 'login' command.

        %(prog)s registry.example.org

Once you no longer need to authenticate, 'logout' again.

Registry interaction is (currently) handled by 'skopeo', which will store
credentials in `${XDG_RUNTIME_DIR}/containers/auth.json`.
""",
    )
    p.add_argument(
        "-u",
        "--username",
        help="Username for registry",
    )
    p.add_argument(
        "-p",
        "--password",
        help="Password for registry",
    )
    p.add_argument(
        "registry",
        help="Registry to login to (e.g. 'registry.example.org')",
    )

    # ################### logout ###################
    p = add_default_args(
        subparsers,
        "logout",
        func=do_logout,
        help="Logout from a registry",
        formatter_class=formatter,
        epilog="""
For pushing images to a registry (and sometimes for pulling),
you often need to authenticate yourself. Do this via the 'login' command.

Once you no longer need to authenticate, 'logout' again.

        %(prog)s registry.example.org
""",
    )

    p.add_argument(
        "registry",
        help="""Registry to logout from (e.g. 'registry.example.org').
        Use '*' for removing cached credentials for ALL registries in the auth file.""",
    )

    # ################### pull ###################
    p = add_default_args(
        subparsers,
        "pull",
        func=do_pull,
        help="Pull VM from a registry, and import it into libvirt",
        formatter_class=formatter,
        epilog="""
        This is a shortcut for 'sarde fetch' and 'sarde import'
""",
    )

    g = p.add_argument_group("input")
    g.add_argument(
        "container",
        help=f"Container to pull (e.g. {example_container!r})",
    )
    g.add_argument(
        "-d",
        "--ocidir",
        help="""Directory where OCI containers are stored (DEFAULT: %(default)r)""",
    )
    g.add_argument(
        "-s",
        "--source",
        help="Fetch the specified container, rather than <container>",
    )

    g = p.add_argument_group("output")
    # <name> is omitted (or optional) - use the SHA256 instead
    g.add_argument(
        "-n",
        "--name",
        help="Destination VM name (DEFAULT: create a unique name)",
    )
    g.add_argument(
        "-o",
        "--vmdir",
        "--outdir",
        help="Output directory to extract VM disk images to (DEFAULT: %(default)r)",
    )

    g = p.add_argument_group("return")
    g.add_argument(
        "--print-uuid",
        action="store_true",
        help="Return UUID of the imported VM, rather than it's name.",
    )

    # ################### fetch ###################
    p = add_default_args(
        subparsers,
        "fetch",
        func=do_fetch,
        help="Fetch a VM from the registry",
        formatter_class=formatter,
        epilog="""
Download a VM image from the registry, and store it in the OCI cache
(as specified via the '--ocidir' flag).

        %(prog)s registry.example.org/sardines/debian:latest

This will fetch the given image and store it as a number of blobs under
'${ocidir}/registry.example.org/sardines/debian/'.
Use 'import' to create the actual VM from these blobs.
""",
    )
    g = p.add_argument_group("input")
    g.add_argument(
        "-s",
        "--source",
        help="Fetch the specified container, rather than <container>",
    )
    g = p.add_argument_group("output")
    g.add_argument(
        "container",
        help=f"Container to fetch (e.g. {example_container!r})",
    )
    g.add_argument(
        "-d",
        "--ocidir",
        help="""Directory where OCI containers are stored (DEFAULT: %(default)r)""",
    )

    # ################### push ###################
    p = add_default_args(
        subparsers,
        "push",
        func=do_push,
        help="Push a VM to a registry",
        formatter_class=formatter,
        epilog="""Push a container from the local OCI cache (e.g. previously 'export'ed from a VM),
to a remote registry.
Most likely you will have to 'login' first, in order to be allowed to push.
By default, the local container name is used to determine the destination.
E.g. a container named 'registry.example.org/sardines/debian:stable' will be pushed to the 'registry.example.org'
host as 'sardines/debian:stable'.

        %(prog)s registry.example.org/sardines/debian:stable

You can specify a different destination with the '--target' flag.

        %(prog)s --target 'registry.example.org/sardines/debian:latest' 'registry.example.org/sardines/debian:bookworm'

This is especially useful, if the local OCI container does not point to a registry:
        %(prog)s --target 'registry.example.org/sardines/debian:latest' 'debian:bookworm'
""",
    )
    g = p.add_argument_group("input")
    g.add_argument(
        "container",
        help=f"Container to push (e.g. {example_container!r})",
    )
    g = p.add_argument_group("output")
    g.add_argument(
        "-d",
        "--ocidir",
        help="""Directory where OCI containers are stored (DEFAULT: %(default)r)""",
    )
    g.add_argument(
        "-t",
        "--target",
        help="Push to the specified container, rather than <container>",
    )

    # ################### list ###################
    p = add_default_args(
        subparsers,
        "list",
        func=do_list,
        help="List local containers.",
        formatter_class=formatter,
        epilog="""
List all available containers, found in the local OCI cache.
Apart from the source (currently always 'oci') and the name of the image
(as specified when 'fetch'ing or 'clone'ing an image), it will also print
the image ID (shared by identical containers) and the current state,
which can be either "missing" (if the container has not yet been 'import'ed),
"stopped" (if the container has been imported but is not running) or
"running"" (if the container is currently a running as a VM).

        $ %(prog)s
        SOURCE  NAME                                                   IMAGE ID               STATE
        oci     registry.example.org/sardines/debian@sha256:f99c5178   sha256:052bd348        stopped
        oci     registry.example.org/sardines/debian:latest            sha256:052bd348           "
        oci     foo.example.com/ubuntu@sha256:c44c1d53                 sha256:f2c4820d        missing

A machine-friendly output can be obtained via the '--format json' flag.
""",
    )

    g = p.add_argument_group("input")
    g.add_argument(
        "-d",
        "--ocidir",
        help="""Directory where OCI containers are stored (DEFAULT: %(default)r)""",
    )
    g = p.add_argument_group("input")
    g.add_argument(
        "--format",
        default="text",
        choices=["text", "json"],
        type=lambda x: str(x).lower(),
        help="""Directory where OCI containers are stored (DEFAULT: %(default)r)""",
    )

    # ################### getconfig ###################
    p = add_default_args(
        subparsers,
        "getconfig",
        func=do_getconfig,
        help="Get VM config from an OCI bundle",
        formatter_class=formatter,
        epilog="""Gets the (very basic) configuration of the OCI container,
detailing users (and their passwords) as well as exposed services;
formatted as JSON.

        %(prog)s registry.example.org/sardines/debian:latest
        {
          "users": [
            {
              "username": "vagrant",
              "password": "vagrant"
            }
          ],
          "services": [
            {
              "service": "ssh",
              "port": 22,
              "protocol": "tcp"
            }
          ]
        }

The configuration might be empty, in which case, nothing is printed.
""",
    )

    g = p.add_argument_group("input")
    g.add_argument(
        "container",
        help="""Container created by 'sarde export' or 'sarde fetch' (found in <OCIDIR>).""",
    )
    gx = g.add_mutually_exclusive_group()
    gx.add_argument(
        "-d",
        "--ocidir",
        help="""Directory where OCI containers are stored (DEFAULT: %(default)r)""",
    )
    gx.add_argument(
        "--raw",
        action="store_true",
        help="""Assume that <container> is a directory path, rather than an OCI container within <OCIDIR>""",
    )

    # ################### export ###################
    p = add_default_args(
        subparsers,
        "export",
        func=do_export,
        help="Export VM to an OCI bundle",
        formatter_class=formatter,
        epilog="""
Convert a VM disk image to an OCI container

This will create an OCI-container 'mydebian:latest' from the 'mydebian' VM:
        %(prog)s --chunksize 256MiB mydebian mydebian:latest

This should create a (local) OCI-container 'registry.example.com/sardines/mydebian:sid' from the 'mydebian' VM:
        %(prog)s --chunksize 256MiB mydebian 'registry.example.com/sardines/mydebian:sid

Once a VM has been exported to an OCI-container, you can 'push' it to a registry.


NOTE
If your VM uses a disk with a backing file, that backing file MUST be in the same directory
as the depending disk image, and it MUST use a relative path.
To make a backing file relative, you can use something like:

        qemu-img rebase -F qcow2 -u -b base.img child.img

""",
    )

    g = p.add_argument_group("input")
    g.add_argument(
        "--vmconfig",
        help="""JSON-file containing additional VM configuration (like username/password).""",
    )
    g.add_argument(
        "--libvirtconfig",
        help="""XML-file containing the libvirt configuration for the VM.
        Useful if you want to export the disks of an existing VM with a generic VM configuration.
        The DEFAULT is to obtain configuration from libvirt.
        Normally you shouldn't need this (EXPERTS ONLY).""",
    )
    g.add_argument(
        "name",
        help="""Source VM name""",
    )

    g = p.add_argument_group("output")
    g.add_argument(
        "container",
        nargs="?",
        help="""Container to export into.""",
    )
    g.add_argument(
        "--chunksize",
        type=parse_size,
        help="""Split gzipped disk images into chunks of the given size
        (in bytes; but specifiers like 'GiB' or 'MB' are recognized). (DEFAULT: do not split)""",
    )
    gx = g.add_mutually_exclusive_group()
    gx.add_argument(
        "-d",
        "--ocidir",
        help="""Directory where OCI containers are stored (DEFAULT: %(default)r)""",
    )
    gx.add_argument(
        "--raw",
        action="store_true",
        help="""Assume that <container> is a directory path, rather than an OCI container within <OCIDIR>""",
    )

    # ################### import ###################
    p = add_default_args(
        subparsers,
        "import",
        func=do_import,
        help="Import VM from an OCI bundle",
        formatter_class=formatter,
        epilog="""Create a VM from an OCI container.

OCI containers are not runnable per se, you have to first re-assemble them into VM (disk images,...).
This is done via the 'import' command.

The following will create a VM named 'mydebian' from the local 'registry.example.org/sardines/debian:latest' container.

        %(prog)s -n mydebian registry.example.org/sardines/debian:latest

The container has to be available locally.
A convenient alternative is to use the 'pull' command, which will fetch the container first if needed.
""",
    )

    g = p.add_argument_group("input")
    g.add_argument(
        "container",
        help="""Container created by 'sarde export' or 'sarde fetch' (found in <OCIDIR>).""",
    )
    gx = g.add_mutually_exclusive_group()
    gx.add_argument(
        "-d",
        "--ocidir",
        help="""Directory where OCI containers are stored (DEFAULT: %(default)r)""",
    )
    gx.add_argument(
        "--raw",
        action="store_true",
        help="""Assume that <container> is a directory path, rather than an OCI container within <OCIDIR>""",
    )

    g = p.add_argument_group("output")
    # <name> is omitted (or optional) - use the SHA256 instead
    g.add_argument(
        "-n",
        "--name",
        help="Destination VM name (DEFAULT: create a unique name)",
    )
    g.add_argument(
        "-o",
        "--vmdir",
        "--outdir",
        help="Output directory to extract VM disk images to (DEFAULT: %(default)r)",
    )

    g = p.add_argument_group("return")
    g.add_argument(
        "--print-uuid",
        action="store_true",
        help="Return UUID of the imported VM, rather than it's name.",
    )

    # ################### clone ###################
    p = add_default_args(
        subparsers,
        "clone",
        func=do_clone,
        help="Clone a VM",
        description="Creates a local virtual machine by cloning either a remote or another local virtual machine.",
        formatter_class=formatter,
        epilog="""
""",
    )

    g = p.add_argument_group("input")
    g.add_argument(
        "-d",
        "--ocidir",
        help="""Directory where OCI containers are stored (DEFAULT: %(default)r)""",
    )
    g.add_argument(
        "container",
        metavar="source",
        help="""Source container.""",
    )

    g = p.add_argument_group("output")
    g.add_argument(
        "-o",
        "--vmdir",
        "--outdir",
        help="Output directory to extract VM disk images to (DEFAULT: %(default)r)",
    )
    g.add_argument(
        "target",
        help="""Target VM.""",
    )

    # ################### selftest ###################
    p = add_default_args(
        subparsers,
        "selftest",
        func=do_selftest,
        help="Test logging with verbosity and whatelse",
        formatter_class=formatter,
        epilog="""
        '%(prog)s' creates output at various loglevels and prints some internal data about the Python state.
""",
    )

    # ################### ###### ###################

    common_args, remainder = common.parse_known_args()
    args = parser.parse_args()

    # fix common args
    for k, v in common_args._get_kwargs():
        if v is None:
            continue
        setattr(args, k, v)

    # merge verbosity settings
    if args.verbose or args.quiet:
        args.verbosity = (args.verbose or 0) - (args.quiet or 0)
        verbosity = args.verbosity
    else:
        args.verbosity = None
        verbosity = 0  # defaults.verbosity
    # logging.setVerbosity(verbosity)
    log.setLevel(max(1, logging.INFO - 10 * verbosity))
    del args.quiet
    del args.verbose

    fun = args.func
    del args.func

    return (fun, args)


def _main():
    userconf = os.getenv("SARDE_CONF_FILES")
    if userconf:
        configfiles = userconf.split(":")
    else:
        configfiles = [
            "/etc/sardinecake/sarde.ini",
            "~/.config/sardinecake/sarde.ini",
        ]
    defaults = readConfig(configfiles, with_default=True)

    try:
        fun, args = parseArgs(defaults)
    except SystemExit as e:
        if e.code and type(e.code) != int:
            log.fatal(e)
        raise

    log.debug(f"configuration-files: {configfiles}")
    log.debug(f"default config     : {defaults}")
    log.debug(f"function           : {fun}")
    log.debug(f"arguments          : {args}")

    if args.no_run:
        raise SystemExit(0)

    ret = None
    try:
        ret = fun(args)
    except SystemExit as e:
        if e.code and type(e.code) != int:
            log.fatal(e)
            raise SystemExit(SYSTEM_FAILURE)
        raise (e)
    raise SystemExit(ret)


if __name__ == "__main__":
    _main()



#!/usr/bin/env python3

# gitlab-sardinecake-executor - run GitLab-CI jobs within libvirt VMs
#
# Copyright © 2024, IOhannes m zmölnig, forum::für::umläute
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

_name = "sardinecake.defaults"
_version = None

import logging

from .namespace import Namespace

try:
    from ._version import version as _version
except ImportError:
    pass

log = logging.getLogger(_name)

sarde = Namespace(
    oci_path="./oci",
    vmdisk_path="./",
)

executor = Namespace(
    username="vagrant",
    password="vagrant",
    timeout=30,
    verbosity=0,
)


BUILD_FAILURE = 1
SYSTEM_FAILURE = 2


def _main():
    logging.basicConfig()


if __name__ == "__main__":
    _main()



#!/usr/bin/env python3

# sardinecake - run GitLab-CI jobs within libvirt VMs
#
# Copyright © 2024, IOhannes m zmölnig, forum::für::umläute
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import subprocess

log = logging.getLogger("sardinecake.virsh")


class virSH:
    cmd_context = {
        "autostart": "domain",
        "backup-begin": "domain",
        "backup-dumpxml": "domain",
        "blkiotune": "domain",
        "checkpoint-create": "domain",
        "checkpoint-create-as": "domain",
        "checkpoint-list": "domain",
        "console": "domain",
        "cpu-baseline": "file",
        "cpu-compare": "file",
        "cpu-stats": "domain",
        "create": "file",
        "define": "file",
        "desc": "domain",
        "destroy": "domain",
        "domblkerror": "domain",
        "domblkinfo": "domain",
        "domblklist": "domain",
        "domblkstat": "domain",
        "domcontrol": "domain",
        "domdirtyrate-calc": "domain",
        "domdisplay": "domain",
        "domdisplay-reload": "domain",
        "domfsfreeze": "domain",
        "domfsinfo": "domain",
        "domfsthaw": "domain",
        "domfstrim": "domain",
        "domhostname": "domain",
        "domid": "domain",
        "domifaddr": "domain",
        "domiflist": "domain",
        "dominfo": "domain",
        "domjobabort": "domain",
        "domjobinfo": "domain",
        "domlaunchsecinfo": "domain",
        "dommemstat": "domain",
        "domname": "domain",
        "dompmwakeup": "domain",
        "domsetlaunchsecstate": "domain",
        "domstate": "domain",
        "domtime": "domain",
        "domuuid": "domain",
        "dump": "domain",
        "dumpxml": "domain",
        "edit": "domain",
        "emulatorpin": "domain",
        "guest-agent-timeout": "domain",
        "guestinfo": "domain",
        "guestvcpus": "domain",
        "hypervisor-cpu-compare": "file",
        "iface-define": "file",
        "iface-destroy": "interface",
        "iface-dumpxml": "interface",
        "iface-edit": "interface",
        "iface-mac": "interface",
        "iface-name": "interface",
        "iface-start": "interface",
        "iface-undefine": "interface",
        "inject-nmi": "domain",
        "iothreadinfo": "domain",
        "lxc-enter-namespace": "domain",
        "managedsave": "domain",
        "managedsave-dumpxml": "domain",
        "managedsave-edit": "domain",
        "managedsave-remove": "domain",
        "memtune": "domain",
        "metadata": "domain",
        "migrate": "domain",
        "migrate-compcache": "domain",
        "migrate-getmaxdowntime": "domain",
        "migrate-getspeed": "domain",
        "migrate-postcopy": "domain",
        "net-autostart": "network",
        "net-create": "file",
        "net-define": "file",
        "net-desc": "network",
        "net-destroy": "network",
        "net-dhcp-leases": "network",
        "net-dumpxml": "network",
        "net-edit": "network",
        "net-info": "network",
        "net-name": "network",
        "net-port-list": "network",
        "net-start": "network",
        "net-undefine": "network",
        "net-uuid": "network",
        "nodedev-autostart": "device",
        "nodedev-create": "file",
        "nodedev-define": "file",
        "nodedev-destroy": "device",
        "nodedev-detach": "device",
        "nodedev-dumpxml": "device",
        "nodedev-info": "device",
        "nodedev-reattach": "device",
        "nodedev-reset": "device",
        "nodedev-start": "device",
        "nodedev-undefine": "device",
        "numatune": "domain",
        "nwfilter-binding-create": "file",
        "nwfilter-define": "file",
        "perf": "domain",
        "pool-autostart": "pool",
        "pool-build": "pool",
        "pool-create": "file",
        "pool-define": "file",
        "pool-delete": "pool",
        "pool-destroy": "pool",
        "pool-dumpxml": "pool",
        "pool-edit": "pool",
        "pool-info": "pool",
        "pool-name": "pool",
        "pool-refresh": "pool",
        "pool-start": "pool",
        "pool-undefine": "pool",
        "pool-uuid": "pool",
        "qemu-agent-command": "domain",
        "qemu-monitor-command": "domain",
        "reboot": "domain",
        "reset": "domain",
        "restore": "file",
        "resume": "domain",
        "save": "domain",
        "save-image-dumpxml": "file",
        "save-image-edit": "file",
        "schedinfo": "domain",
        "screenshot": "domain",
        "secret-define": "file",
        "secret-dumpxml": "secret",
        "secret-get-value": "secret",
        "secret-set-value": "secret",
        "secret-undefine": "secret",
        "send-key": "domain",
        "shutdown": "domain",
        "snapshot-create": "domain",
        "snapshot-create-as": "domain",
        "snapshot-current": "domain",
        "snapshot-delete": "domain",
        "snapshot-edit": "domain",
        "snapshot-info": "domain",
        "snapshot-list": "domain",
        "snapshot-parent": "domain",
        "snapshot-revert": "domain",
        "start": "domain",
        "suspend": "domain",
        "ttyconsole": "domain",
        "undefine": "domain",
        "update-memory-device": "domain",
        "vcpucount": "domain",
        "vcpuinfo": "domain",
        "vcpupin": "domain",
        "vncdisplay": "domain",
        "vol-delete": "vol",
        "vol-dumpxml": "vol",
        "vol-info": "vol",
        "vol-key": "vol",
        "vol-list": "pool",
        "vol-name": "vol",
        "vol-path": "vol",
        "vol-pool": "vol",
        "vol-wipe": "vol",
    }

    def __init__(self, context, libvirtURI=None):
        self._context = context
        self._cmd = ["virsh"]
        if libvirtURI:
            self._cmd += ["--connect", libvirtURI]

    def __call__(self, cmd: str, *args, **kwargs):
        ctx = self.cmd_context.get(cmd)
        if ctx and self._context:
            context = [f"--{ctx}", self._context]
        else:
            context = []

        command = self._cmd + [cmd] + context + list(args)
        kwargs.setdefault("stderr", subprocess.DEVNULL)

        # set LANG to 'C' (unless overridden)
        env = kwargs.get("env")
        if env and "LANG" in env:
            pass
        else:
            if not env:
                env = dict(os.environ)
            env["LANG"] = "C"
        kwargs["env"] = env
        return subprocess.run(command, **kwargs)

    def isRunning(self) -> bool | None:
        """check if VM is running (->true), or not (->false)
        returns None if the VM does not exist"""
        try:
            x = self("domstate", check=True, stdout=subprocess.PIPE)
        except:
            return
        if b"running" in x.stdout:
            return True
        return False

    def domifaddr(self) -> dict:
        """get IPs of VM"""
        x = self("domifaddr", stdout=subprocess.PIPE)
        try:
            x.check_returncode()
        except:
            return

        stdout = [line for line in x.stdout.splitlines()[2:] if line]
        if not stdout:
            return

        ret = {}
        for line in stdout:
            data = line.decode().split()
            ip = data[3].split("/")[0]
            ret[data[0]] = {
                "mac": data[1],
                "protocol": data[2],
                "address": ip,
                "network": data[3],
            }
        return ret

    def destroy(self, graceful=False, remove_logs=False) -> bool:
        """destroy domain"""
        args = []
        if graceful:
            args += ["--graceful"]
        if remove_logs:
            args += ["--remove-logs"]
        x = self("destroy", *args, stdout=subprocess.PIPE)
        return x.returncode == 0


if __name__ == "__main__":
    import argparse

    logging.basicConfig()

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--connect",
        type=str,
        help="connection URI for libvirt",
    )
    parser.add_argument(
        "vm",
        help="VM name",
    )
    args = parser.parse_args()
    if args.connect is None:
        virsh = virSH(args.vm)
    else:
        virsh = virSH(args.vm, libvirtURI=args.connect)
    print(virsh)
    print(f"running: {virsh.isRunning()}")
    print(f"IP: {virsh.domifaddr()}")



#!/usr/bin/env python3

# gitlab-sardinecake-executor - run GitLab-CI jobs within libvirt VMs
#
# Copyright © 2024, IOhannes m zmölnig, forum::für::umläute
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

_name = "sardinecake.executor"
_version = None

import fnmatch
import os
import time

from . import logger as logging
from . import clone, defaults, ssh
from .importvm import getconfig
from .namespace import Namespace
from .virsh import virSH

try:
    from ._version import version as _version
except ImportError:
    pass


log = logging.getLogger(_name)

BUILD_FAILURE = 1
SYSTEM_FAILURE = 2
BUILD_EXIT_CODE_FILE = None

custom_env = {}


def setup():
    global BUILD_FAILURE
    global SYSTEM_FAILURE
    global BUILD_EXIT_CODE_FILE
    global custom_env

    BUILD_FAILURE = int(os.environ.get("BUILD_FAILURE_EXIT_CODE", BUILD_FAILURE))
    SYSTEM_FAILURE = int(os.environ.get("SYSTEM_FAILURE_EXIT_CODE", SYSTEM_FAILURE))
    BUILD_EXIT_CODE_FILE = os.environ.get("BUILD_EXIT_CODE_FILE", BUILD_EXIT_CODE_FILE)

    prefix = "CUSTOM_ENV_"
    custom_env = {
        k[len(prefix) :]: v for k, v in os.environ.items() if k.startswith(prefix)
    }

    # fallback for easier testing
    for k in ["CI_JOB_IMAGE", "CI_JOB_ID"]:
        if k in custom_env:
            continue
        v = os.environ.get(k)
        if v:
            custom_env[k] = v


def getVersion(libvirtURI=None):
    """get the version of the executor
    (possibly including libvirt version)
    """

    def parseVersion(ver):
        major = ver // 1000000
        minor = (ver % 1000000) // 1000
        release = ver % 1000
        return (major, minor, release)

    try:
        import libvirt

        with libvirt.openReadOnly(libvirtURI) as conn:
            typ = conn.getType()
            libver = {
                "libvirt": parseVersion(conn.getLibVersion()),
                typ: parseVersion(conn.getVersion()),
            }
            libvirt_version = ", ".join(
                f"{t}: {v0}.{v1}.{v2}" for t, (v0, v1, v2) in libver.items()
            )

    except Exception as e:
        log.debug("Unable to query libvirt for version", exc_info=True)
        libvirt_version = None

    ver = ""
    if _version:
        ver = _version
    if libvirt_version:
        ver = f"{ver} ({libvirt_version})"
    ver = ver.strip()
    if ver:
        return ver.strip()


def getVMimage() -> str | None:
    if not custom_env:
        setup()
    env = "CI_JOB_IMAGE"
    img = custom_env.get(env)
    if not img:
        log.fatal(f"no VM defined via {env!r}")
        return None
    return img


def getVMname() -> str | None:
    if not custom_env:
        setup()
    env = "CI_JOB_ID"
    job = custom_env.get(env)
    if job:
        return f"gitlab-{job}"
    log.fatal(f"no GitlabEnv {env!r} found")
    return None


def getVMaddress(
    name: str | None = None,
    timeout: float | None = None,
    libvirtURI: str | None = None,
    protocol: str = "IPv4",
) -> str | None:
    if not name:
        name = getVMname()
    proto = None
    if protocol:
        proto = protocol.lower()

    virsh = virSH(name, libvirtURI=libvirtURI)

    if virsh.isRunning() is None:
        log.error(f"VM {name!r} does not exist")
        return None

    start = time.time() or 0
    now = start
    timeout = timeout or 0

    data = None
    while int(now - start) <= timeout:
        try:
            data = virsh.domifaddr()
            if data:
                break
        except:
            log.debug(f"failed getting {protocol} address for {name!r}", exc_info=True)

        time.sleep(0.5)
        if timeout is not None:
            now = time.time() or 0

    if not data:
        return None

    for dev, ifconf in data.items():
        if proto and ifconf.get("protocol").lower() != proto:
            continue
        IP = ifconf.get("address")
        if IP:
            log.debug(
                f"VM {name!r} has {protocol} {IP} (after {time.time() - start} seconds)"
            )
            return IP
    return None


def checkVMOnline(
    vmname: str,
    username: str | None = None,
    password: str | None = None,
    key_filename: str | list[str] | None = None,
    port: int | None = None,
    timeout: float | None = None,
    libvirtURI: str | None = None,
) -> str | None:
    """check if the given VM can be accessed with ssh://username:password@...
    on success, returns the IP of the VM; otherwise returns None
    """
    try:
        log.info("Waiting for the VM to become online...")
        IP = getVMaddress(vmname, timeout=timeout, libvirtURI=libvirtURI)
    except Exception as e:
        log.exception(f"failed to get IP address for {vmname!r}")
        return None

    if not IP:
        log.exception(f"couldn't get IP address for {vmname!r}")
        return None

    if username:
        # check whether the host is SSH-able
        log.info("Waiting for the VM to become SSH-able...")
        if not ssh.checkSSH(
            IP,
            username=username,
            password=password,
            key_filename=key_filename,
            port=port,
            timeout=timeout,
        ):
            return None
        log.info("Was able to SSH!")
        log.debug(f"ssh://{username}@{IP} success!")

    log.info("VM is ready.")
    return IP


def getAuth(args: Namespace, vmconfig: dict = {}) -> dict:
    """get dict with authentication information from args"""
    auth = {}
    vmconfig = vmconfig or {}

    users = vmconfig.get("users")
    if users:
        u = users[0] or {}
        auth.setdefault("username", u.get("username"))
        auth.setdefault("password", u.get("password"))
        auth.setdefault("port", u.get("port"))
    auth = {k: v for k, v in auth.items() if v is not None}

    try:
        auth.setdefault("username", args.username)
    except AttributeError:
        pass
    try:
        auth.setdefault("password", args.password)
    except AttributeError:
        pass

    try:
        keypath = args.identity_file
        if keypath:
            if os.path.isdir(keypath):
                # list files in keypath (https://stackoverflow.com/a/3207973/1169096)
                keypath = next(os.walk(keypath), (None, None, []))[2]
            auth.setdefault("key_filename", keypath)
    except AttributeError:
        pass

    return auth


def do_logtest(args):
    """print at various loglevels"""
    for name, level in logging.getLevelNamesMapping().items():
        log.log(level, name)

    print(f"log-level: {log.level}/{log.getEffectiveLevel()}")
    print(args)


def do_config(args):
    """configure subsequent stages of GitLab runner the runner"""
    import json
    import sys

    driver = {
        "name": _name or "gitlab-sardinecake-executor",
    }
    ver = getVersion(args.connect)
    if ver:
        driver["version"] = ver

    try:
        builds_dir = args.guest_builds_dir
    except AttributeError:
        pass
    try:
        cache_dir = args.guest_cache_dir
    except AttributeError:
        pass

    job_env = {}

    # build the config dictionary
    localvars = locals()

    def makeConfig(*args):
        result = {}
        for k in args:
            v = localvars.get(k)
            if v:
                result[k] = v
        return result

    data = makeConfig(
        "driver",
        "job_env",
        "builds_dir",
        "cache_dir",
    )

    json.dump(data, sys.stdout)


def do_prepare(args):
    """create a VM gitlab-${CI_JOB_ID} and wait till it is accessible via SSH"""
    baseVM = None
    cloneVM = None
    timeout = args.timeout
    libvirtURI = args.connect
    outdir = args.vmdir
    ocidir = args.ocidir
    clonedir = args.clonedir

    if outdir and not os.path.exists(outdir):
        # make sure that 'outdir' points to a directory
        outdir = os.path.join(outdir, "")

    if not custom_env:
        setup()

    shownbase = baseVM = custom_env.get("CI_JOB_IMAGE")
    if args.image and args.image != baseVM:
        shownbase = f"{baseVM!r} ({args.image!r})"
        baseVM = args.image

    if not baseVM:
        raise SystemExit("no base image given")

    cloneVM = getVMname()
    if not cloneVM:
        raise SystemExit("unable to get name of cloned VM")

    # start VM
    log.info(f"Pulling the latest version of {shownbase!r} as {cloneVM!r}...")
    try:
        cfg = clone.fetch_and_clone(
            container=baseVM,
            vm=cloneVM,
            ocidir=ocidir,
            vmdir=outdir,
            clonedir=clonedir,
            startmode=clone.START_EPHEMERAL,
            libvirtURI=libvirtURI,
        )
    except Exception as e:
        exc_info = log.getEffectiveLevel() < logging.INFO
        log.fatal(e, exc_info=exc_info)
        return SYSTEM_FAILURE

    auth = getAuth(args, cfg)

    # now wait until we have an SSH-connection
    host = checkVMOnline(
        cloneVM,
        timeout=timeout,
        libvirtURI=libvirtURI,
        **auth,
    )
    if not host:
        return SYSTEM_FAILURE

    if args.prepare_script:
        log.info(f"Running prepare script {args.prepare_script!r}")
        try:
            with open(args.prepare_script) as f:
                cmds = f.read()
            ret = ssh.run_script(
                host,
                args.prepare_script,
                env=custom_env,
                **auth,
            )
            if ret:
                log.error(f"prepare script returned {ret}")
                do_cleanup(args)
                return BUILD_FAILURE
        except Exception as e:
            log.exception("failed to run prepare script")
            do_cleanup(args)
            return BUILD_FAILURE

    user = auth.get("username")
    log.debug(f"{cloneVM!r} is ready as {host!r}")
    if user:
        log.debug(f"You can login via SSH as {user}@{host}")


def do_run(args):
    """run a script within the VM"""
    timeout = 30
    libvirtURI = args.connect

    vmname = getVMname()

    container = custom_env.get("CI_JOB_IMAGE")
    cfg = getconfig(os.path.join(args.ocidir, container))

    auth = getAuth(args, cfg)

    if args.action:
        log.debug(f"Running {args.action!r}")
    else:
        log.warning("Running unknown action")

    if not vmname:
        raise SystemExit("unable to get name of cloned VM")
    try:
        host = getVMaddress(vmname, timeout=timeout, libvirtURI=libvirtURI)
    except Exception as e:
        log.exception(f"failed to get IP address for {vmname!r}")
        return False

    if not host:
        log.fatal(f"Unable to get address of {vmname!r}")
        return SYSTEM_FAILURE
    try:
        ret = ssh.run_script(host, args.script, **auth)
        if ret:
            log.error(f"{args.action} script returned {ret}")
            if BUILD_EXIT_CODE_FILE:
                try:
                    with open(BUILD_EXIT_CODE_FILE, "w") as f:
                        f.write(f"{ret}\n")
                except Exception as e:
                    log.error(f"failed to write {BUILD_EXIT_CODE_FILE!r}: {e}")
            return BUILD_FAILURE
    except Exception as e:
        log.exception(f"failed to run {args.action} script")
        return BUILD_FAILURE
    return ret


def do_cleanup(args):
    """cleanup the VM"""
    cloneVM = getVMname()

    if not cloneVM:
        raise SystemExit("unable to get name of cloned VM")

    virsh = virSH(cloneVM, libvirtURI=args.connect)

    log.info(f"destroy VM {cloneVM!r}")
    if not virsh.destroy():
        raise SystemExit(f"couldn't destroy VM {cloneVM!r}")


def parseArgs():
    import argparse

    def add_common_args(parser):
        p = parser.add_argument_group("libvirt options")
        p.add_argument(
            "--connect",
            help="libvirt connection URI",
        )

        p = parser.add_argument_group("verbosity")
        p.add_argument(
            "-v",
            "--verbose",
            action="count",
            help="raise verbosity (can be given multiple times)",
        )
        p.add_argument(
            "-q",
            "--quiet",
            action="count",
            help="lower verbosity (can be given multiple times)",
        )

        p.add_argument(
            "--no-run",
            action="store_true",
            help="don't actually run (exit early)",
        )

        p = parser.add_argument_group(
            "SSH options",
            description="fallback SSH credentials (if none are provided by the VM)",
        )
        p.add_argument(
            "-u",
            "--username",
            help=f"SSH user (DEFAULT: {defaults.executor.username})",
        )
        p.add_argument(
            "-p",
            "--password",
            help="SSH password",
        )
        p.add_argument(
            "--identity-file",
            help="SSH identity file for authentication (if a password is required, uses '--password')",
        )

        return parser

    def add_default_args(subparsers, name=None, fallbacks={}, func=None, **kwargs):
        if any([name, func]) and not all([subparsers, name, func]):
            raise ValueError("inconsistent arguments for add_subparser()")

        if name:
            if "help" in kwargs and "description" not in kwargs:
                kwargs["description"] = kwargs["help"]
            p = subparsers.add_parser(name, **kwargs)
            add_common_args(p)
        else:
            p = subparsers

        # fallbacks.setdefault("ocidir", defaults.sarde.oci_path)
        # fallbacks.setdefault("vmdir", defaults.sarde.vmdisk_path)
        # fallbacks.setdefault("verbose", 0)
        # fallbacks.setdefault("quiet", 0)
        mydefaults = defaults.executor.get(name) or defaults.executor.get(None) or {}
        fallbacks.update(mydefaults)
        fallbacks["func"] = func

        if func:
            p.set_defaults(**fallbacks)

        return p

    # common args for all sub-parsers and the top-level parser
    common = argparse.ArgumentParser(add_help=False)
    add_common_args(common)

    parser = argparse.ArgumentParser(
        description="""Custom GitLab Runner executor to run jobs inside ephemeral libvirt virtual machines.""",
        parents=[
            common,
        ],
    )

    add_default_args(parser)

    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=_version or "<unknown>",
    )

    subparsers = parser.add_subparsers(required=True, title="commands")
    p = add_default_args(
        subparsers, "config", func=do_config, help="Configure GitLab Runner"
    )

    p = add_default_args(
        subparsers,
        "prepare",
        func=do_prepare,
        help="Prepare a libvirt VM for execution",
    )
    p.add_argument(
        "--timeout",
        type=int,
        help=f"libvirt connection timeout for VM boot and SSH connection (DEFAULT: {defaults.executor.timeout})",
    )
    p.add_argument(
        "--prepare-script",
        help="path to script (on hypervisor) that is executed within the VM to complete the preparation",
    )
    p.add_argument(
        "--clonedir",
        help="directory for storing disk images of (ephemeral) clone VMs",
    )
    g = p.add_argument_group("expert/dev options")
    g.add_argument(
        "--vmdir",
        help="directory where VM images (extracted from the OCI containers) are stored",
    )
    g.add_argument(
        "--ocidir",
        help="directory where OCI containers are stored",
    )
    p.add_argument(
        "--omit-backingstore",
        default=None,
        action="store_true",
        help="Do not create a <backingStore/> tag in the new guest definition (might be needed for older libvirt versions)",
    )

    p = add_default_args(
        subparsers, "run", func=do_run, help="Run GitLab's scripts in libvirt VM"
    )
    p.add_argument(
        "script",
        help="script file to execute",
    )
    p.add_argument(
        "action",
        help="Name of the action being executed",
    )

    p = add_default_args(
        subparsers,
        "cleanup",
        func=do_cleanup,
        help="Cleanup libvirt VM after job finishes",
    )

    common_args, remainder = common.parse_known_args()
    args = parser.parse_args()

    # fix common args
    for k, v in common_args._get_kwargs():
        if v is None:
            continue
        setattr(args, k, v)

    # merge verbosity settings
    if args.verbose or args.quiet:
        args.verbosity = (args.verbose or 0) - (args.quiet or 0)
        verbosity = args.verbosity
    else:
        args.verbosity = None
        verbosity = defaults.executor.verbosity
    log.setLevel(max(1, logging.INFO - 10 * verbosity))

    del args.quiet
    del args.verbose

    fun = args.func
    del args.func

    return (fun, args)


def parseConfigs(configfiles, default):
    """parse configuration file(s)
    any given keyword arguments override settings in the configuration files
    """
    import configparser

    # general options
    # - libvirt URI
    # - specify VM clone/start command
    # - only allow images that have a configuration section
    # per VM options
    # - ssh credentials
    # - cache/build directories (currently not implemented)
    example = """
[DEFAULT]
# general options
connect = qemu:///system

# VM-defaults
username = vagrant
password = vagrant
guest-builds-dir "/path/on/guest/builds"
guest-cache-dir "/path/on/guest/cache"
builds-dir "/path/on/host/builds"
cache-dir "/path/on/host/cache"


# VM overrides (per VM-name)
[win10]
    username = admin
    password = secret
    """

    def section2dict(section, options=None, excluded_options=[]) -> dict:
        """turn a ConfigParser section into a dict"""
        getters = {
            "omit_backingstore": "boolean",
            "timeout": "float",
            "verbosity": "int",
        }
        result = {}
        if options is None:
            options = cfg.options(section)
        for option in options:
            opt = option.lower().replace("-", "_")
            if opt in excluded_options:
                continue
            getter = getattr(cfg, "get" + getters.get(opt, ""))
            result[opt] = getter(section, option)
        return result

    globaloptions = ["verbosity", "no-run"]
    default_configfiles = [
        "/etc/gitlab-runner/gitlab-sardinecake-executor.ini",
        os.path.expanduser("~/.config/gitlab-runner/gitlab-sardinecake-executor.ini"),
    ]

    cfg = configparser.ConfigParser()
    default_section = cfg.default_section
    cfg.default_section = None
    log.debug(f"Loading configurations from {configfiles or default_configfiles}")
    did_configfiles = cfg.read(configfiles or default_configfiles)
    if not did_configfiles:
        log.warning(f"No configuration files found!")
    else:
        log.debug(f"Loaded configurations from {did_configfiles}")

    data = {}
    for section in cfg:
        if section is None:
            continue
        sectiondata = section2dict(section, excluded_options=globaloptions)
        if sectiondata:
            data[section] = sectiondata

    if default_section in data:
        defaultcfg = data[default_section]
        del data[default_section]
    else:
        defaultcfg = {}

    defaultcfg.update(default)
    data[None] = defaultcfg

    return data


def getConfiguration() -> dict:
    """get configurations.
    this is the configuration precedence (first beats the rest):
    - container specific config
      - container-specific config in .ini-files
      - container-specific config from embedded config
    - general config
      - cmdline arguments
      - default section in .ini-files
    *this* function does not know about the embedded configs yet.
    so it returns a dictionary with VM -> config mappings.
    the special VM <None> contains the general config (merged)
    """

    img = getVMimage()

    # read cmdline arguments
    fun, args = parseArgs()
    log.debug(args)
    cfg_default = {k: v for k, v in args.__dict__.items() if v is not None}

    # read configuration
    configfiles = []
    userconf = os.getenv("GITLAB_SARDINECAKE_EXECUTOR_CONF_FILES")
    if userconf:
        configfiles = userconf.split(":")
    else:
        try:
            configfiles = args.config
            del args.config
        except AttributeError:
            pass

    cfg = parseConfigs(configfiles, cfg_default)
    log.debug(cfg)

    # get the default configuration for our image:
    cfg_default = cfg[None]
    del cfg[None]
    cfg_image = {}
    if img in cfg:
        cfg_image = cfg[img]
    elif img:
        wildcardconfigs = {k: v for k, v in cfg.items() if fnmatch.fnmatch(img, k)}
        if wildcardconfigs:
            # get the best match (most beginning characters match):
            cfg_image = sorted(
                wildcardconfigs.items(), key=lambda x: len(x[0]), reverse=True
            )[0][1]

    imgcfg = Namespace(**cfg_image)
    imgcfg.setdefaults(cfg_default)
    imgcfg.setdefaults(defaults.executor)

    log.debug(imgcfg)
    if imgcfg.verbosity is not None:
        log.setLevel(max(1, logging.INFO - 10 * imgcfg.verbosity))

    return (fun, imgcfg)


def _main():
    global log
    logging.basicConfig()
    log = logging.getLogger("sardinecake")
    setup()
    try:
        fun, args = getConfiguration()
    except SystemExit as e:
        if e.code and type(e.code) != int:
            log.fatal(e)
            raise SystemExit(SYSTEM_FAILURE)
        raise (e)

    if args.no_run:
        raise SystemExit(0)

    ret = None
    try:
        ret = fun(args)
    except SystemExit as e:
        if e.code and type(e.code) != int:
            log.fatal(e)
            raise SystemExit(SYSTEM_FAILURE)
        raise (e)
    raise SystemExit(ret)


if __name__ == "__main__":
    _main()



#!/usr/bin/env python3

# sardinecake - manage VMs via OCI
#
# Copyright © 2024, IOhannes m zmölnig, forum::für::umläute
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


# import a libvirt VM into an OCI bundle

import hashlib
import json
import logging
import os

import libvirt


try:
    from .domain import Domain
    from . import gziparchive as gz
    from . import qcow
    from . import utils
except ImportError:
    from domain import Domain
    import gziparchive as gz
    import qcow
    import utils


log = logging.getLogger("sardinecake.importvm")


def digest_to_path(digest: str, ocidir: str = "") -> str:
    """convert a digest (e.g. 'sha256:12345') into a path <ocidir>/blobs/<hashtype>/<digest>"""
    pathcomponents = [ocidir, "blobs"]
    pathcomponents += digest.split(":", maxsplit=1)
    return os.path.join(*pathcomponents)


def check_manifest(m: dict, tag: str | None = None, ocidir: str = "") -> str:
    """check if manifest <m> as found in the OCIindex is usable; returns the path to the manifest"""
    digest = None
    if issubclass(type(tag), (list, tuple)):
        digest = ":".join(tag)
        tag = None

    if digest and m.get("digest") != digest:
        return
    if tag and m.get("annotations", {}).get("org.opencontainers.image.ref.name") != tag:
        return
    manifest_path = digest_to_path(m["digest"], ocidir=ocidir)
    if os.path.exists(manifest_path):
        return manifest_path
    log.error(f"manifest not found at {manifest_path!r}")


def getOCIindex(ocidir: str) -> dict:
    """get OCI-index (which includes information on all tags)"""
    # read metadata
    try:
        with open(os.path.join(ocidir, "oci-layout")) as f:
            oci_layout = json.load(f)
        with open(os.path.join(ocidir, "index.json")) as f:
            index = json.load(f)
    except FileNotFoundError as e:
        log.fatal(f"{ocidir} does not look like an OCI-compliant directory: {e}")
        return
    except json.decoder.JSONDecodeError as e:
        log.fatal(f"invalid JSON data in {ocidir!r}: {e}")
        return
    return index


def getOCImanifest(tag: str | None, index: dict, ocidir: str) -> dict:
    """get the manifest for the given tag from the <index> (as returned by getOCIindex()"""
    manifests = index.get("manifests")
    if not manifests:
        log.fatal("no manifests in the container")
        return

    manifest_info = None
    manifest_path = None
    for m in reversed(manifests):
        manifest_path = check_manifest(m, tag=tag or "latest", ocidir=ocidir)
        if manifest_path:
            manifest_info = m
            break

    if not manifest_info and not tag:
        # if the user did not request a specific tag, and there's no 'latest' manifest,
        # just pick the one added last
        m = manifests[-1]
        manifest_path = check_manifest(m, tag=None, ocidir=ocidir)
        if manifest_path:
            manifest_info = m

    if not manifest_info:
        log.fatal("no usable manifest found")
        return

    try:
        with open(manifest_path) as f:
            manifest = json.load(f)
    except OSError:
        log.fatal(f"could not read manifest {manifest_path!r}")
        return
    except json.decoder.JSONDecodeError as e:
        log.fatal(f"invalid JSON data in manifest {manifest_path}: {e}")
        return

    return manifest


def getOCIconfig(manifest: dict, ocidir: str) -> tuple[str, dict]:
    """get the image configuration from a manifest (as returned by getOCImanifest)"""
    config = manifest["config"]
    oci_config_name = config["digest"]
    oci_config = digest_to_path(oci_config_name, ocidir=ocidir)

    if config["mediaType"] != "application/vnd.oci.image.config.v1+json":
        log.fatal(f"unknown mediaType for config: {config['mediaType']!r}")
        return

    try:
        with open(oci_config) as f:
            oci_config = json.load(f)
    except OSError:
        log.fatal(f"could not read oci-config {oci_config!r}")
        return
    except json.decoder.JSONDecodeError as e:
        log.fatal(f"invalid JSON data in oci-config {oci_config!r}: {e}")
        return

    return (oci_config_name, oci_config)


def getOCIlayers(manifest: dict, ocidir: str) -> dict[str, list[str]]:
    """get the layers from the manifest in a mediaType -> [paths] dictionary.
    returns None if there's an error (like a non-existing layer file
    """
    layers = {}
    for lay in manifest.get("layers", []):
        laytype = lay.get("mediaType").split(";")[0]
        laypath = digest_to_path(lay.get("digest"), ocidir=ocidir)
        if not os.path.exists(laypath):
            log.fatal(f"layer {laypath!r} does not exist")
            return
        layers[laytype] = layers.get(laytype, []) + [laypath]
    return layers


def _importVM(
    domain: Domain,
    outdir=".",
    libvirtURI: str | None = None,
    gzipdata=[str],
    quiet: bool = False,
) -> str | None:
    """import a libvirt VM from an OCI compatible bundle.
    returns (name, UUID) of the newly created VM (or None)
    """

    if quiet:
        libvirt.registerErrorHandler((lambda ctx, err: 1), None)

    outdir = os.path.abspath(outdir)
    disks = domain.getDiskFiles()
    log.debug(f"disks: {disks}")
    absdisks = {
        k: v if os.path.isabs(v) else os.path.join(outdir, v) for k, v in disks.items()
    }
    log.debug(f"-> absdisks: {absdisks}")
    domain.changeDiskFiles(absdisks)

    name = domain.getName()

    uuid = None
    with libvirt.open(libvirtURI) as conn:
        olddom = Domain.fromLibVirt(conn, name)
        if olddom:
            log.info(f"VM {name} already exists")
            return (olddom.getName(), olddom.getUUID())
        try:
            x = conn.defineXML(str(domain))
            uuid = (x.name(), x.UUIDString())
        except:
            log.exception(f"couldn't create VM {name}")
            return

    if gzipdata:
        gzipdir = os.path.split(gzipdata[0])[0]
        gzipchunks = [os.path.split(_)[1] for _ in gzipdata]
        try:
            gz.gunzipchunked(gzipdir, outdir, gzipchunks)
        except:
            log.exception(f"failed to extract data into {outdir!r}")
            raise SystemExit("failed to extract data")

    return uuid


def importvm(
    ocidir: str,
    outdir: str | None = "./",
    name: str | bool = False,
    libvirtURI: str | None = None,
    quiet: bool = False,
) -> str | None:
    """import a libvirt VM from an OCI compatible bundle.

    if <name> is True, it uses the original name of the VM.
    if <name> is False (or falsish otherwise,  e.g. ''), it uses the SHA256 of the OCI-bundle as the name
    if <outdir> ends with "/", the files are put into a VMnamed subdirectory therein.

    returns the path to the bundle (a subdirectory of <outdir>).
    or None.
    """
    # after a successful run we have:
    """
    <outdir>/<name>/disk1.qemu
    <outdir>/<name>/disk2.qemu
    """
    # resp
    """
    <outdir>/disk1.qemu
    <outdir>/disk2.qemu
    """
    if "/" in (str(name)):
        raise ValueError(f"invalid VM name {name!r}")

    if not outdir:
        outdir = "./"

    ocidir, tag = utils.split_imagetag(ocidir)

    # read metadata
    index = getOCIindex(ocidir)
    if not index:
        return
    manifest = getOCImanifest(tag, index, ocidir)
    if not manifest:
        return

    config_type = manifest.get("config", {}).get("mediaType")
    if config_type != "application/vnd.oci.image.config.v1+json":
        log.fatal(
            f"manifest {manifest_path!r} has invalid OCI image configuration of type {config_type!r}"
        )
        return

    try:
        oci_config_name, oci_config = getOCIconfig(manifest, ocidir)
    except TypeError:
        return

    layers = getOCIlayers(manifest, ocidir)

    domain = None
    libvirt_config = layers.get("application/vnd.libvirt.domain.config+xml")
    if libvirt_config:
        domain = Domain.fromXMLFile(libvirt_config[0])

    if not domain:
        log.fatal(f"couldn't read VM information from {libvirt_config}")
        return

    gzipdata = []
    for k, v in layers.items():
        if k.endswith("+gzip"):
            gzipdata += v

    if name is True:
        name = domain.getName()
        log.debug(f"use original name {name!r}")
    elif not name:
        name = oci_config_name
        domain.changeName(name)
        log.debug(f"use digest name {name!r}")
    else:
        domain.changeName(name)
        log.debug(f"use explicit name {name!r}")

    if not os.path.split(outdir)[1]:
        outdir = os.path.join(outdir, name)

    log.debug(f"importing VM {name!r} to {outdir!r} using chunks {gzipdata}")

    os.makedirs(outdir, exist_ok=True)
    uuid = _importVM(
        domain, outdir=outdir, libvirtURI=libvirtURI, gzipdata=gzipdata, quiet=quiet
    )
    return uuid


def getconfig(ocidir: str) -> dict:
    """get sardinecake configuration from the container at <ocidir>"""
    ocidir, tag = utils.split_imagetag(ocidir)

    # read metadata
    index = getOCIindex(ocidir)
    if not index:
        return
    manifest = getOCImanifest(tag, index, ocidir)
    if not manifest:
        return

    config_type = manifest.get("config", {}).get("mediaType")
    if config_type != "application/vnd.oci.image.config.v1+json":
        log.fatal(
            f"manifest {manifest_path!r} has invalid OCI image configuration of type {config_type!r}"
        )
        return

    try:
        oci_config_name, oci_config = getOCIconfig(manifest, ocidir)
    except TypeError:
        return

    layers = getOCIlayers(manifest, ocidir)
    if layers is None:
        return
    config = layers.get("application/vnd.sardinecake.vm.config.v1+json")
    if not config:
        return
    for c in config:
        with open(c) as f:
            return json.load(f)


def _main():
    import argparse

    p = argparse.ArgumentParser(
        description="import a libvirt VM from an OCI-compliant directory. (such as created by 'sarde export')",
    )
    p.add_argument(
        "--connect",
        help="libvirt connection URI",
    )
    p.add_argument(
        "-n",
        "--name",
        help="VM name to create (DEFAULT: the sha256 of the OCI); specify an empty value ('') to use the original VM name",
    )

    p.add_argument(
        "-o",
        "--outdir",
        default="./",
        help="directory to create VM-files in (use a trailing slash to create a VMnamed directory within)",
    )

    p.add_argument(
        "ocidir",
        help="OCI-compliant directory to import VM from",
    )

    args = p.parse_args()

    x = importvm(
        args.ocidir, outdir=args.outdir, name=args.name, libvirtURI=args.connect
    )
    print(x)


if __name__ == "__main__":
    _main()



#!/usr/bin/env python3

# sardinecake - manage VMs via OCI
#
# Copyright © 2024, IOhannes m zmölnig, forum::für::umläute
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import logging
import subprocess

import libvirt

try:
    from .domain import Domain
    from . import qcow
except ImportError:
    from domain import Domain
    import qcow

log = logging.getLogger("sardinecake.clone")

START_PRINTXML = 0
START_DEFINE = 1
START_EPHEMERAL = 2
START_PERSISTENT = 3


def cloneQCOW2(source: str, target: str | None = None) -> str:
    """copy-on-write clone of a QCOW2 file `source` to another QCOW2 file.

    - if target is `None`, the clone will be in the same directory as the source, and the name will be derived from the source basename.
    - if target is a directory (or ends with "/"), the clone will be in the given directory, and the name will be derived from the source basename.
    - otherwise, target is a the prospective name of the output file.

    `cloneQCOW2()` will refuse to overwrite existing images, and use a unique filename when required.
    the actual filename of the cloned image is returned
    """
    import subprocess

    # qemu-img create -f qcow2 -b base.qcow2 -F qcow2 clone.qcow2

    # check if source exists and can be opened (otherwise raise a standard error)
    with open(source) as f:
        pass

    if not target:
        target = source
    elif os.path.isdir(target) or not os.path.basename(target):
        target = os.path.join(target, os.path.basename(source))

    # ensure that output directory exists
    outdir = os.path.dirname(target)
    os.makedirs(outdir, exist_ok=True)

    base, ext = os.path.splitext(target)

    i = ""
    while True:
        target = f"{base}{i}{ext}"
        try:
            targetfd = open(target, "x")
            break
        except FileExistsError:
            pass
        if not i:
            i = 0
        i -= 1

    targetfd.close()

    log.debug(f"shallow-cloning {source!r} to {target!r}")

    q = qcow.QCOW(target)
    q.create(backing_file=source)

    if not q:
        return

    return target


def cloneDisks(disks: dict[str, str], outputdir: str | None) -> dict[str, str]:
    """shallow-clones the given disks into output dir.

    `disks` is a dictionary that maps IDs to disk paths.
    a similar map from IDs to cloned paths is returned
    """
    return {k: cloneQCOW2(v, outputdir) for k, v in disks.items()}


def clone(
    srcname: str,
    dstname: str,
    startmode: int = START_DEFINE,
    libvirtURI: str | None = None,
    mac: str | None = None,
    outdir: str | None = None,
    quiet: bool = False,
    omit_backingstore: bool | None = None,
):
    """do a fast clone of a VM

    - `srcname`: original VM
    - `dstname`: VM to be created
    - `startmode`: how to start the newly created VM
    - `libvirtURI`: connection URI to the libvirt server
        (note that only local QCOW2 disks can be cloned)
    - `mac`: MAC-address of cloned network interface (use None for a new random MAC)
    - `outdir`: place for the shallow cloned disks images (use None to use the same directory as the input disk images)
    - `quiet`: if `True`, suppress libvirt errors
    - `omit_backingstore`:
        - if `False`, creates a `<backingStore/>` entry for COW-images
        - if `True`, does not create a `<backingStore/>` entry for COW-images (might work around a bug with older libvirt)
        - if `None`, tries to autodetect whether the `<backingStore/>` should be omitted
    """

    if quiet:
        libvirt.registerErrorHandler((lambda ctx, err: 1), None)

    libvirt_open = libvirt.open
    if startmode == START_PRINTXML:
        libvirt.openReadOnly

    with libvirt_open(libvirtURI) as conn:
        if omit_backingstore is None:
            omit_backingstore = conn.getLibVersion() < 9010000

        dom = Domain.fromLibVirt(conn, srcname)
        if not dom:
            raise KeyError(f"Couldn't find original VM {srcname!r}")

        if not dstname:
            suffix = 0
            while True:
                if suffix:
                    name = f"{srcname}-clone-{suffix}"
                else:
                    name = f"{srcname}-clone"
                if not Domain.fromLibVirt(conn, name):
                    dstname = name
                    break
                suffix += 1

        if Domain.fromLibVirt(conn, dstname):
            raise KeyError(f"cloned VM {dstname!r} already exists")

        dom.anonymize(mac_address=mac)
        dom.changeName(dstname)

        srcdisks = dom.getClonableDisks()

        log.debug(f"cloning disks {srcdisks} to {outdir!r}")
        cloneddisks = cloneDisks(srcdisks, outdir)
        log.debug(f"cloned disks {cloneddisks}")

        backing_files = {}
        if not omit_backingstore:
            backing_files = {
                dev: qcow.getBackingChain(disk) for dev, disk in srcdisks.items()
            }
        dom.changeDiskFiles(cloneddisks, backing_files=backing_files)

        if startmode == START_PRINTXML:
            print(dom)
        elif startmode == START_EPHEMERAL:
            log.debug(f"start ephemeral VM {dstname!r}")
            conn.createXML(dom.toXML())
            for d in cloneddisks.values():
                log.debug(f"deleting ephemeral volume {d!r}")
                os.unlink(d)
        else:
            log.debug(f"define cloned VM {dstname!r}")
            vm = conn.defineXML(dom.toXML())
            if startmode == START_PERSISTENT:
                log.debug(f"starting persistent VM {dstname!r}")
                vm.create()


def fetch_and_clone(
    container: str,
    vm: str,
    ocidir: str,
    vmdir: str,
    clonedir: str | None = None,
    startmode: int = START_DEFINE,
    libvirtURI: str | None = None,
):
    """fetches a 'container', imports it and creates an ephemeral clone.

    - returns the configuration of the VM (as known).
    - raises exceptions on failure
    """
    from .ociregistry import fetch
    from .importvm import importvm, getconfig

    if ocidir is None:
        oci = container
    else:
        oci = os.path.join(ocidir or "", container)

    clonedir = clonedir or vmdir

    quiet = (log.getEffectiveLevel() >= logging.INFO,)

    log.info(f"fetching {container!r}")
    if not fetch(container, ocidir=ocidir):
        raise RuntimeError(f"failed to fetch {container!r}")

    log.info(f"importing {container!r}")
    basevm = importvm(
        ocidir=oci,
        outdir=vmdir,
        name=None,
        libvirtURI=libvirtURI,
        quiet=quiet,
    )
    if not basevm:
        raise RuntimeError(f"failed to import {container}")

    config = getconfig(ocidir=oci) or {}
    config["name"] = vm

    log.info(f"cloning {container!r} as {vm}")
    clone(
        srcname=basevm[0],
        dstname=vm,
        startmode=startmode,
        libvirtURI=libvirtURI,
        outdir=clonedir,
        quiet=quiet,
        omit_backingstore=None,
    )

    return config


# #########################################################################
def _parseArgs():
    """parse commandline arguments"""
    import argparse

    parser = argparse.ArgumentParser(
        description="""Duplicate a virtual machine, changing all the unique host side configuration like MAC address, name, etc.""",
        epilog="""This is very similar to 'virt-clone(1)', but uses shallow clones for copy-on-write disks.
Currently, only QCOW2 disks are supported.
    """,
    )
    parser.set_defaults(start=START_DEFINE)

    parser.add_argument(
        "--connect",
        metavar="URI",
        help="Connect to hypervisor with libvirt URI (note that to-be-cloned disks must be locally available)",
    )

    parser.add_argument(
        "name",
        metavar="NEW_NAME",
        nargs="?",
        help="Name for the new guest (overrides the '--name' argument)",
    )

    g = parser.add_argument_group("General Options")
    g.add_argument(
        "-o",
        "--original",
        metavar="SRC_NAME",
        required=True,
        help="Name of the original guest to clone.",
    )
    g.add_argument(
        "-n",
        "--name",
        dest="opt_name",
        metavar="NEW_NAME",
        help="Name for the new guest",
    )

    g = parser.add_argument_group("Clone Options")
    g.add_argument(
        "-m",
        "--mac",
        metavar="NEW_MAC",
        help="New fixed MAC address for the clone guest. Default is a randomly generated MAC.",
    )

    g = parser.add_argument_group("Output Options")
    g.add_argument(
        "--outdir",
        help="Directory to put cloned disk images into",
    )

    g.add_argument(
        "--omit-backingstore",
        action="store_true",
        default=None,
        help="Do not create a <backingStore/> tag in the new guest definition (might be needed for older libvirt versions)",
    )
    g = parser.add_mutually_exclusive_group()
    g.add_argument(
        "--define",
        dest="start",
        action="store_const",
        const=START_DEFINE,
        help="Define a new persistent VM. (DEFAULT)",
    )
    g.add_argument(
        "--print-xml",
        dest="start",
        action="store_const",
        const=START_PRINTXML,
        help="Print the generated domain XML rather than create the guest.",
    )
    g.add_argument(
        "--start-ephemeral",
        dest="start",
        action="store_const",
        const=START_EPHEMERAL,
        help="Start the cloned VM as an ephemeral domain (the VM and all it's disks will be destroyed once powered down).",
    )
    g.add_argument(
        "--start-persistent",
        dest="start",
        action="store_const",
        const=START_PERSISTENT,
        help="Define a new persistent VM and start it.",
    )

    g = parser.add_argument_group("Verbosity")
    g.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="Suppress non-error output",
    )
    g.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="Print debugging information",
    )

    args = parser.parse_args()

    if not args.name:
        args.name = args.opt_name
    del args.opt_name

    # make sure that outdir is a directory (not a filename)
    if args.outdir is not None:
        args.outdir = os.path.join(args.outdir, "")

    if args.debug:
        log.setLevel(logging.DEBUG)

    return args


def _main():
    logging.basicConfig()
    args = _parseArgs()
    log.debug(args)

    try:
        clone(
            args.original,
            args.name,
            startmode=args.start,
            libvirtURI=args.connect,
            mac=args.mac,
            outdir=args.outdir,
            quiet=args.quiet,
            omit_backingstore=args.omit_backingstore,
        )
    except Exception as e:
        log.debug("clone failed", exc_info=True)
        raise SystemExit(e)


if __name__ == "__main__":
    _main()



#!/usr/bin/env python3

# sardinecake - manage VMs via OCI
#
# Copyright © 2024, IOhannes m zmölnig, forum::für::umläute
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


__version__ = "<unknown>"
__author__ = "IOhannes m zmölnig <zmoelnig@iem.at>"

from .auth import login, logout
from .exportvm import exportvm
from .importvm import importvm
from .ociregistry import fetch, push

try:
    from ._version import version as __version__
except ImportError:
    pass

__all__ = ["login", "logout", "exportvm", "importvm", "fetch", "push", "Domain", "QCOW"]
